# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import TypeVar

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def transpose(tensor: Tensor) -> Tensor:
    """Transpose the `tensor`.

    This method can be used with Numpy data:
    ```python
    n = np.array([[0,1,2],[3,4,5],[6,7,8]])
    b = fe.backend.transpose(n)  # [[0, 3, 6], [1, 4, 7], [2, 5, 8]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([[0,1,2],[3,4,5],[6,7,8]])
    b = fe.backend.transpose(t)  # [[0, 3, 6], [1, 4, 7], [2, 5, 8]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([[0,1,2],[3,4,5],[6,7,8]])
    b = fe.backend.transpose(p)  # [[0, 3, 6], [1, 4, 7], [2, 5, 8]]
    ```

    Args:
        tensor: The input value.

    Returns:
        The transposed `tensor`.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        return tf.transpose(tensor)
    elif isinstance(tensor, torch.Tensor):
        return tensor.T
    elif isinstance(tensor, np.ndarray):
        return np.transpose(tensor)
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import shutil
import tempfile
import unittest
from typing import Union

import pydot
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.architecture.pytorch import LeNet as PyLeNet
from fastestimator.architecture.tensorflow import LeNet
from fastestimator.dataset.data import mnist
from fastestimator.op.numpyop.univariate import ExpandDims, Minmax
from fastestimator.op.tensorop.loss import CrossEntropy
from fastestimator.op.tensorop.model import ModelOp, UpdateOp
from fastestimator.schedule import cosine_decay
from fastestimator.trace.adapt import LRScheduler
from fastestimator.trace.io import Traceability
from fastestimator.trace.metric import Accuracy
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import FeInputSpec


def _lacks_graphviz():
    try:
        pydot.Dot.create(pydot.Dot())
    except OSError:
        return True
    return False


def _build_estimator(model: Union[tf.keras.Model, torch.nn.Module], trace: Traceability, axis: int = -1):
    train_data, eval_data = mnist.load_data()
    test_data = eval_data.split(0.5)
    batch_size = 32
    pipeline = fe.Pipeline(train_data=train_data,
                           eval_data=eval_data,
                           test_data=test_data,
                           batch_size=batch_size,
                           ops=[ExpandDims(inputs="x", outputs="x", axis=axis), Minmax(inputs="x", outputs="x")])
    network = fe.Network(ops=[
        ModelOp(model=model, inputs="x", outputs="y_pred"),
        CrossEntropy(inputs=("y_pred", "y"), outputs="ce"),
        UpdateOp(model=model, loss_name="ce")
    ])
    traces = [
        Accuracy(true_key="y", pred_key="y_pred"),
        LRScheduler(model=model, lr_fn=lambda step: cosine_decay(step, cycle_length=3750, init_lr=1e-3)),
        trace
    ]
    estimator = fe.Estimator(pipeline=pipeline,
                             network=network,
                             epochs=1,
                             traces=traces,
                             max_train_steps_per_epoch=1,
                             max_eval_steps_per_epoch=None)
    fake_data = tf.ones(shape=(batch_size, 28, 28, 1)) if axis == -1 else torch.ones(size=(batch_size, 1, 28, 28))
    model.fe_input_spec = FeInputSpec(fake_data, model)
    return estimator


class TestTraceability(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.root_dir = os.path.join(tempfile.gettempdir(), "FEUnitTestReports")
        cls.tf_dir = os.path.join(cls.root_dir, "TF")
        cls.torch_dir = os.path.join(cls.root_dir, "Torch")

    @unittest.skipIf(_lacks_graphviz(), "The machine does not have GraphViz installed")
    def test_tf_traceability(self):
        if os.path.exists(self.tf_dir) and os.path.isdir(self.tf_dir):
            shutil.rmtree(self.tf_dir)

        trace = Traceability(save_path=self.tf_dir)
        est = _build_estimator(fe.build(model_fn=LeNet, optimizer_fn="adam", model_name='tfLeNet'), trace)

        trace.system = est.system
        trace.system.epoch_idx = 1
        trace.system.summary.name = "TF Test"

        trace.on_begin(Data())
        trace.on_end(Data())

        crawler = os.walk(self.tf_dir)
        root = next(crawler)
        self.assertIn('resources', root[1], "A resources subdirectory should have been generated")
        self.assertIn('tf_test.tex', root[2], "The tex file should have been generated")
        # Might be a pdf and/or a .ds_store file depending on system, but shouldn't be more than that
        self.assertLessEqual(len(root[2]), 3, "Extra files should not have been generated")
        figs = next(crawler)
        self.assertIn('tf_test_tfLeNet.pdf', figs[2], "A figure for the model should have been generated")
        self.assertIn('tf_test_logs.png', figs[2], "A log image should have been generated")
        self.assertIn('tf_test.txt', figs[2], "A raw log file should have been generated")

    @unittest.skipIf(_lacks_graphviz(), "The machine does not have GraphViz installed")
    def test_torch_traceability(self):
        if os.path.exists(self.torch_dir) and os.path.isdir(self.torch_dir):
            shutil.rmtree(self.torch_dir)

        trace = Traceability(save_path=self.torch_dir)
        est = _build_estimator(fe.build(model_fn=PyLeNet, optimizer_fn="adam", model_name='torchLeNet'), trace, axis=0)

        trace.system = est.system
        trace.system.epoch_idx = 1
        trace.system.summary.name = "Torch Test"

        trace.on_begin(Data())
        trace.on_end(Data())

        crawler = os.walk(self.torch_dir)
        root = next(crawler)
        self.assertIn('resources', root[1], "A figures subdirectory should have been generated")
        self.assertIn('torch_test.tex', root[2], "The tex file should have been generated")
        # Might be a pdf and/or a .ds_store file depending on system, but shouldn't be more than that
        self.assertLessEqual(len(root[2]), 3, "Extra files should not have been generated")
        figs = next(crawler)
        self.assertIn('torch_test_torchLeNet.pdf', figs[2], "A figure for the model should have been generated")
        self.assertIn('torch_test_logs.png', figs[2], "A log image should have been generated")
        self.assertIn('torch_test.txt', figs[2], "A raw log file should have been generated")



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

from fastestimator.summary import average_summaries, Summary, ValWithError


class TestSummary(unittest.TestCase):
    def test_merge(self):
        summary1 = Summary(name='test1')
        summary1.history['train']['acc'] = {50: 0.9}

        summary2 = Summary(name='test2')
        summary2.history['train']['acc'] = {80: 0.8}

        summary1.merge(summary2)

        self.assertEqual(summary1.history['train']['acc'][50], 0.9)
        self.assertEqual(summary1.history['train']['acc'][80], 0.8)


class TestAverageSummaries(unittest.TestCase):
    def test_empty_summaries(self):
        result = average_summaries(name="ex1", summaries=[])
        with self.subTest("Should return a summary instance"):
            self.assertTrue(isinstance(result, Summary))
        with self.subTest("Name should be correct"):
            self.assertEqual("ex1", result.name)

    def test_single_summary(self):
        s1 = Summary('ex2')
        s1.history['train']['acc'][10] = 0.55
        result = average_summaries(name="ex1", summaries=[s1])
        with self.subTest("Should return a summary instance"):
            self.assertTrue(isinstance(result, Summary))
        with self.subTest("Name should be correct"):
            self.assertEqual("ex1", result.name)
        with self.subTest("History should be correct"):
            self.assertEqual(0.55, result.history['train']['acc'][10])

    def test_multi_summaries(self):
        s1 = Summary('s1')
        s2 = Summary('s2')

        s1.history['train']['acc'] = {0: 0.2, 10: 0.4, 20: 0.7, 45: ValWithError(0.8, 0.9, 1.0)}
        s2.history['train']['acc'] = {0: 0.3, 10: 0.4, 20: 0.9, 30: 1.0}

        s1.history['test']['mcc'] = {45: 0.834}

        s1.history['eval']['wombats'] = {5: 4, 10: 9}
        s2.history['eval']['wombats'] = {5: '3 wombats', 10: '7 wombats'}

        s_merge = average_summaries('s', [s1, s2])

        with self.subTest("Should return a summary instance"):
            self.assertTrue(isinstance(s_merge, Summary))
        with self.subTest("Name should be correct"):
            self.assertEqual("s", s_merge.name)
        with self.subTest("Paired datapoints should be matched"):
            self.assertEqual({0, 10, 20, 30, 45}, s_merge.history['train']['acc'].keys())
        with self.subTest("Paired datapoints should have correct mean and std"):
            self.assertEqual(0.25, s_merge.history['train']['acc'][0].y)
            self.assertEqual(0.17929, round(s_merge.history['train']['acc'][0].y_min, 5))
            self.assertEqual(0.32071, round(s_merge.history['train']['acc'][0].y_max, 5))
        with self.subTest("Paired datapoints with zero std should be handled correctly"):
            self.assertEqual(0.4, s_merge.history['train']['acc'][10])
        with self.subTest("Partially paired datapoints should be handled correctly"):
            self.assertEqual(0.9, s_merge.history['train']['acc'][45])
            self.assertEqual(1.0, s_merge.history['train']['acc'][30])
        with self.subTest("Unpaired datapoints should be handled correctly"):
            self.assertEqual(0.834, s_merge.history['test']['mcc'][45])
        with self.subTest("String values should be handled correctly"):
            self.assertEqual(3.5, s_merge.history['eval']['wombats'][5].y)
            self.assertEqual(2.79289, round(s_merge.history['eval']['wombats'][5].y_min, 5))
            self.assertEqual(8.0, s_merge.history['eval']['wombats'][10].y)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest
from io import StringIO
from unittest.mock import patch

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace import Logger
from fastestimator.util.data import Data


class TestLogger(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Data({})
        cls.on_begin_global_step_msg = "FastEstimator-Start: step: 2;"
        cls.on_begin_msg = "FastEstimator-Start: step: 1;"
        cls.on_batch_end_msg = "FastEstimator-Train: step: 1;"
        cls.on_epoch_end_train_msg = "FastEstimator-Train: step: 2; epoch: 0;"
        cls.on_epoch_end_eval_msg = "FastEstimator-Eval: step: 2; epoch: 0;"
        cls.on_epoch_end_test_msg = "FastEstimator-Test: step: 2; epoch: 0;"
        cls.on_end_msg = "FastEstimator-Finish: step: 2;"

    def _test_print_msg(self, func, data, msg):
        with patch('sys.stdout', new=StringIO()) as fake_stdout:
            func(data)
            log = fake_stdout.getvalue().strip()
            self.assertEqual(log, msg)

    def test_on_begin_global_step(self):
        logger = Logger()
        logger.system = sample_system_object()
        logger.system.global_step = 2
        self._test_print_msg(func=logger.on_begin, data=self.data, msg=self.on_begin_global_step_msg)

    def test_on_begin_global_step_not_none(self):
        logger = Logger()
        logger.system = sample_system_object()
        self._test_print_msg(func=logger.on_begin, data=self.data, msg=self.on_begin_msg)

    def test_on_batch_end(self):
        logger = Logger()
        logger.system = sample_system_object()
        logger.system.global_step = 1
        logger.system.log_steps = 3
        self._test_print_msg(func=logger.on_batch_end, data=self.data, msg=self.on_batch_end_msg)

    def test_on_epoch_end_mode_train(self):
        logger = Logger()
        logger.system = sample_system_object()
        logger.system.global_step = 2
        logger.system.log_steps = 3
        self._test_print_msg(func=logger.on_epoch_end, data=self.data, msg=self.on_epoch_end_train_msg)

    def test_on_epoch_end_mode_eval(self):
        logger = Logger()
        logger.system = sample_system_object()
        logger.system.mode = 'eval'
        logger.system.global_step = 2
        logger.system.log_steps = 3
        self._test_print_msg(func=logger.on_epoch_end, data=self.data, msg=self.on_epoch_end_eval_msg)

    def test_on_epoch_end_mode_test(self):
        logger = Logger()
        logger.system = sample_system_object()
        logger.system.mode = 'test'
        logger.system.global_step = 2
        logger.system.log_steps = 3
        self._test_print_msg(func=logger.on_epoch_end, data=self.data, msg=self.on_epoch_end_test_msg)

    def test_on_end(self):
        logger = Logger()
        logger.system = sample_system_object()
        logger.system.global_step = 2
        logger.system.log_steps = 3
        self._test_print_msg(func=logger.on_end, data=self.data, msg=self.on_end_msg)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Tuple

import tensorflow as tf
from tensorflow.python.keras import Sequential, layers


# noinspection PyPep8Naming
def LeNet(input_shape: Tuple[int, int, int] = (28, 28, 1), classes: int = 10) -> tf.keras.Model:
    """A standard LeNet implementation in TensorFlow.

    The LeNet model has 3 convolution layers and 2 dense layers.

    Args:
        input_shape: shape of the input data (height, width, channels).
        classes: The number of outputs the model should generate.

    Raises:
        ValueError: Length of `input_shape` is not 3.
        ValueError: `input_shape`[0] or `input_shape`[1] is smaller than 18.

    Returns:
        A TensorFlow LeNet model.
    """
    _check_input_shape(input_shape)
    model = Sequential()
    model.add(layers.Conv2D(32, (3, 3), activation='relu', input_shape=input_shape))
    model.add(layers.MaxPooling2D((2, 2)))
    model.add(layers.Conv2D(64, (3, 3), activation='relu'))
    model.add(layers.MaxPooling2D((2, 2)))
    model.add(layers.Conv2D(64, (3, 3), activation='relu'))
    model.add(layers.Flatten())
    model.add(layers.Dense(64, activation='relu'))
    model.add(layers.Dense(classes, activation='softmax'))
    return model


def _check_input_shape(input_shape):
    if len(input_shape) != 3:
        raise ValueError("Length of `input_shape` is not 3 (channel, height, width)")

    height, width, _ = input_shape

    if height < 18 or width < 18:
        raise ValueError("Both height and width of input_shape need to not smaller than 18")



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.op.numpyop.univariate import ExpandDims
from fastestimator.test.unittest_util import is_equal


class TestExpandDims(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.single_input = np.array([1, 2, 3, 4])
        cls.single_output = [np.array([[1], [2], [3], [4]])]
        cls.multi_input = [np.array([2, 2]), np.array([0, 1, 2])]
        cls.multi_output = [np.array([[2, 2]]), np.array([[0, 1, 2]])]

    def test_single_input(self):
        op = ExpandDims(axis=-1, inputs='x', outputs='x')
        data = op.forward(data=[np.array([1, 2, 3, 4])], state={})
        self.assertTrue(is_equal(data, self.single_output))

    def test_multi_input(self):
        op = ExpandDims(axis=0, inputs='x', outputs='x')
        data = op.forward(data=self.multi_input, state={})
        self.assertTrue(is_equal(data, self.multi_output))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestGetGradient(unittest.TestCase):
    def test_get_gradient_tf_tensor_higher_order_false(self):
        def check_gradient(x):
            with tf.GradientTape(persistent=True) as tape:
                y = x * x
                with self.subTest("check gradient"):
                    obj1 = fe.backend.get_gradient(target=y, sources=x, tape=tape)  # [2.0, 4.0, 6.0]
                    obj2 = tf.constant([2.0, 4.0, 6.0])
                    self.assertTrue(is_equal(obj1, obj2))

                with self.subTest("check gradient of gradient"):
                    obj1 = fe.backend.get_gradient(target=obj1, sources=x, tape=tape)  # None
                    self.assertTrue(is_equal(obj1, None))

        x = tf.Variable([1.0, 2.0, 3.0])
        strategy = tf.distribute.get_strategy()
        if isinstance(strategy, tf.distribute.MirroredStrategy):
            strategy.run(check_gradient, args=(x, ))
        else:
            check_gradient(x)

    def test_get_gradient_tf_tensor_higher_order_true(self):
        def check_gradient(x):
            with tf.GradientTape(persistent=True) as tape:
                y = x * x
                with self.subTest("check gradient"):
                    obj1 = fe.backend.get_gradient(target=y, sources=x, tape=tape, higher_order=True)  # [2.0, 4.0, 6.0]
                    obj2 = tf.constant([2.0, 4.0, 6.0])
                    self.assertTrue(is_equal(obj1, obj2))

                with self.subTest("check gradient of gradient"):
                    obj1 = fe.backend.get_gradient(target=obj1, sources=x, tape=tape)  # [2.0, 2.0, 2.0]
                    obj2 = tf.constant([2.0, 2.0, 2.0])
                    self.assertTrue(is_equal(obj1, obj2))

        x = tf.Variable([1.0, 2.0, 3.0])
        strategy = tf.distribute.get_strategy()
        if isinstance(strategy, tf.distribute.MirroredStrategy):
            strategy.run(check_gradient, args=(x, ))
        else:
            check_gradient(x)

    def test_get_gradient_torch_tensor_higher_order_false(self):
        x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        y = x * x
        obj1 = fe.backend.get_gradient(target=y, sources=x)  # [2.0, 4.0, 6.0]
        obj2 = torch.tensor([2.0, 4.0, 6.0])
        self.assertTrue(is_equal(obj1, obj2))

    def test_get_gradient_torch_tensor_higher_order_true(self):
        x = torch.tensor([1.0, 2.0, 3.0], requires_grad=True)
        y = x * x
        with self.subTest("check gradient"):
            obj1 = fe.backend.get_gradient(target=y, sources=x, higher_order=True)  # [2.0, 4.0, 6.0]
            obj2 = torch.tensor([2.0, 4.0, 6.0])
            self.assertTrue(is_equal(obj1, obj2))

        with self.subTest("check gradient of gradient"):
            obj1 = fe.backend.get_gradient(target=obj1, sources=x)  # [2.0, 2.0, 2.0]
            obj2 = torch.tensor([2.0, 2.0, 2.0])
            self.assertTrue(is_equal(obj1, obj2))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Callable, Dict, List, Optional, Union

import tensorflow as tf
import torch


def update_model(model: Union[tf.keras.Model, torch.nn.Module],
                 gradients: List[Union[tf.Tensor, torch.Tensor]],
                 defer: bool = False,
                 deferred: Optional[Dict[str, List[Callable[[], None]]]] = None) -> None:
    """Update `model` weights based on a given `gradients`.

    This method can be used with TensorFlow models:
    ```python
    m = fe.build(fe.architecture.tensorflow.LeNet, optimizer_fn="adam")
    x = tf.ones((3, 28, 28, 1))  # (batch, height, width, channels)
    y = tf.constant((1, 0, 1))
    with tf.GradientTape(persistent=True) as tape:
        pred = fe.backend.feed_forward(m, x)  # [[~0.5, ~0.5], [~0.5, ~0.5], [~0.5, ~0.5]]
        loss = fe.backend.sparse_categorical_crossentropy(y_pred=pred, y_true=y)  # ~2.3
        gradients = fe.backend.get_gradient(target=loss, sources=m.trainable_variables, tape=tape)
        fe.backend.update_model(m, gradients=gradients)
    ```

    This method can be used with PyTorch models:
    ```python
    m = fe.build(fe.architecture.pytorch.LeNet, optimizer_fn="adam")
    x = torch.ones((3, 1, 28, 28))  # (batch, channels, height, width)
    y = torch.tensor((1, 0, 1))
    pred = fe.backend.feed_forward(m, x)  # [[~0.5, ~0.5], [~0.5, ~0.5], [~0.5, ~0.5]]
    loss = fe.backend.sparse_categorical_crossentropy(y_pred=pred, y_true=y)  # ~2.3
    gradients = fe.backend.get_gradient(target=loss,
                                        sources=[x for x in m.parameters() if x.requires_grad])

    fe.backend.update_model(m, gradients=gradients)
    ```

    Args:
        model: A neural network instance to update.
        gradients: A list of tensors to update the models.
        defer: If True, then the model update function will be stored into the `deferred` dictionary rather than
            applied immediately.
        deferred: A dictionary in which model update functions are stored.

    Raises:
        ValueError: If `model` is an unacceptable data type.
        AssertionError: If `model` doesn't have `current_optimizer` attribute
        AssertionError: If Pytorch `model.current_optimizer` doesn't have `scaler` attribute
    """
    assert hasattr(model, "current_optimizer"), ("The model needs to have 'current_optimizer' attribute. Please "
                                                 "instantiate the model with fe.build")

    if isinstance(model, tf.keras.Model):
        variables = model.trainable_variables
        if defer:
            deferred.setdefault(model.model_name,
                                []).append(lambda: model.current_optimizer.apply_gradients(zip(gradients, variables)))
        else:
            model.current_optimizer.apply_gradients(zip(gradients, variables))

    elif isinstance(model, torch.nn.Module):
        trainable_params = [p for p in model.parameters() if p.requires_grad]
        for gradient, parameter in zip(gradients, trainable_params):
            if parameter.grad is not None:
                parameter.grad += gradient
            else:
                parameter.grad = gradient.clone()
        if defer:
            # Only need to call once per model since gradients are getting accumulated
            deferred[model.model_name] = [lambda: _torch_step(model.current_optimizer)]
        else:
            _torch_step(model.current_optimizer)

            if deferred:
                deferred.pop(model.model_name)  # Don't need those deferred steps anymore
    else:
        raise ValueError("Unrecognized model instance {}".format(type(model)))


def _torch_step(optimizer: torch.optim.Optimizer) -> None:
    assert hasattr(optimizer, "scaler"), ("Pytorch optimizers need to have 'scaler' attribute, Please use fe.build to "
                                          "compile the model")
    if optimizer.scaler is None:
        optimizer.step()
    else:
        optimizer.scaler.step(optimizer)
        optimizer.scaler.update()
    optimizer.zero_grad()



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import TypeVar

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def matmul(a: Tensor, b: Tensor) -> Tensor:
    """Perform matrix multiplication on `a` and `b`.

    This method can be used with Numpy data:
    ```python
    a = np.array([[0,1,2],[3,4,5]])
    b = np.array([[1],[2],[3]])
    c = fe.backend.matmul(a, b)  # [[8], [26]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    a = tf.constant([[0,1,2],[3,4,5]])
    b = tf.constant([[1],[2],[3]])
    c = fe.backend.matmul(a, b)  # [[8], [26]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    a = torch.tensor([[0,1,2],[3,4,5]])
    b = torch.tensor([[1],[2],[3]])
    c = fe.backend.matmul(a, b)  # [[8], [26]]
    ```

    Args:
        a: The first matrix.
        b: The second matrix.

    Returns:
        The matrix multiplication result of a * b.

    Raises:
        ValueError: If either `a` or `b` are unacceptable or non-matching data types.
    """
    if tf.is_tensor(a) and tf.is_tensor(b):
        return tf.matmul(a, b)
    elif isinstance(a, torch.Tensor) and isinstance(b, torch.Tensor):
        return a.matmul(b)
    elif isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return np.matmul(a, b)
    elif type(a) != type(b):
        raise ValueError(f"Tensor types do not match ({type(a)} and {type(b)})")
    else:
        raise ValueError(f"Unrecognized tensor type ({type(a)} or {type(b)})")



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

from fastestimator.test.unittest_util import TraceRun
from fastestimator.trace.metric import Accuracy


class TestAccuracy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.acc_key = "acc"

    def test_tf_one_hot_label(self):
        trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key)
        batch = {"label": tf.constant([[1, 0, 0], [0, 1, 0]])}  # one-hot
        prediction = {"pred": tf.constant([[1, 2, 3], [0.2, 0.5, 0.3]])}
        sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
        sim.run_trace()
        self.assertEqual(sim.data_on_epoch_end[self.acc_key], 0.5)

    def test_tf_class_index_label(self):
        trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key)
        batch = {"label": tf.constant([0, 1])}  # class index
        prediction = {"pred": tf.constant([[1, 2, 3], [0.2, 0.5, 0.3]])}
        sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
        sim.run_trace()
        self.assertEqual(sim.data_on_epoch_end[self.acc_key], 0.5)

    def test_tf_binary_class(self):
        with self.subTest("from_logit=False"):
            trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key, from_logits=False)
            batch = {"label": tf.constant([0, 1])}
            prediction = {"pred": tf.constant([[0.3], [0.6]])}  # pred > 0.5 => class 1
            sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
            sim.run_trace()
            self.assertEqual(sim.data_on_epoch_end[self.acc_key], 1.0)

        with self.subTest("from_logit=True"):
            trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key, from_logits=True)
            batch = {"label": tf.constant([0, 1])}
            prediction = {"pred": tf.constant([[-1], [1]])}  # 1 / 1 + exp(-pred) > 0.5 => class 1
            sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
            sim.run_trace()
            self.assertEqual(sim.data_on_epoch_end[self.acc_key], 1.0)

    def test_torch_one_hot_label(self):
        trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key)
        batch = {"label": torch.tensor([[1, 0, 0], [0, 1, 0]])}  # one-hot
        prediction = {"pred": torch.tensor([[1, 2, 3], [0.2, 0.5, 0.3]])}
        sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
        sim.run_trace()
        self.assertEqual(sim.data_on_epoch_end[self.acc_key], 0.5)

    def test_torch_class_index_label(self):
        trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key)
        batch = {"label": torch.tensor([0, 1])}  # class index
        prediction = {"pred": torch.tensor([[1, 2, 3], [0.2, 0.5, 0.3]])}
        sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
        sim.run_trace()
        self.assertEqual(sim.data_on_epoch_end[self.acc_key], 0.5)

    def test_torch_binary_class(self):
        with self.subTest("from_logit=False"):
            trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key, from_logits=False)
            batch = {"label": torch.tensor([0, 1])}
            prediction = {"pred": torch.tensor([[0.3], [0.6]])}  # pred > 0.5 => class 1
            sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
            sim.run_trace()
            self.assertEqual(sim.data_on_epoch_end[self.acc_key], 1.0)

        with self.subTest("from_logit=True"):
            trace = Accuracy(true_key="label", pred_key="pred", output_name=self.acc_key, from_logits=True)
            batch = {"label": torch.tensor([0, 1])}
            prediction = {"pred": torch.tensor([[-1], [1]])}  # 1 / 1 + exp(-pred) > 0.5 => class 1
            sim = TraceRun(trace=trace, batch=batch, prediction=prediction)
            sim.run_trace()
            self.assertEqual(sim.data_on_epoch_end[self.acc_key], 1.0)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Callable, Dict, Iterable, List, Union

from fastestimator.op.numpyop.numpyop import NumpyOp
from fastestimator.util.traceability_util import traceable


@traceable()
class Tokenize(NumpyOp):
    """Split the sequences into tokens.

    Tokenize split the document/sequence into tokens and at the same time perform additional operations on tokens if
    defined in the passed function object. By default, tokenize only splits the sequences into tokens.

    Args:
        inputs: Key(s) of sequences to be tokenized.
        outputs: Key(s) of sequences that are tokenized.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        tokenize_fn: Tokenization function object.
        to_lower_case: Whether to convert tokens to lowercase.
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 tokenize_fn: Union[None, Callable[[str], List[str]]] = None,
                 to_lower_case: bool = False) -> None:
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.in_list, self.out_list = True, True
        self.tokenize_fn = tokenize_fn
        self.to_lower_case = to_lower_case

    def forward(self, data: List[str], state: Dict[str, Any]) -> List[List[str]]:
        return [self._apply_tokenization(seq) for seq in data]

    def _apply_tokenization(self, data: str) -> List[str]:
        """Split the sequence into tokens and apply lowercase if `do_lower_case` is set.

        Args:
            data: Input sequence.

        Returns:
            A list of tokens.
        """
        if self.tokenize_fn:
            data = self.tokenize_fn(data)
        else:
            data = data.split()
        if self.to_lower_case:
            data = list(map(lambda x: x.lower(), data))
        return data



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.op.tensorop.loss import MeanSquaredError


class TestMeanSquaredError(unittest.TestCase):
    def test_tf(self):
        tf_true = tf.constant([[0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [1, 0, 0, 0]])
        tf_pred = tf.constant([[0.1, 0.9, 0.05, 0.05], [0.1, 0.2, 0.0, 0.7], [0.0, 0.15, 0.8, 0.05],
                               [1.0, 0.0, 0.0, 0.0]])
        mse = MeanSquaredError(inputs='x', outputs='x')
        output = mse.forward(data=[tf_pred, tf_true], state={})
        self.assertTrue(np.allclose(output.numpy(), 0.014375001))

    def test_torch(self):
        torch_true = torch.tensor([[0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [1, 0, 0, 0]])
        torch_pred = torch.tensor([[0.1, 0.9, 0.05, 0.05], [0.1, 0.2, 0.0, 0.7], [0.0, 0.15, 0.8, 0.05],
                                   [1.0, 0.0, 0.0, 0.0]])
        mse = MeanSquaredError(inputs='x', outputs='x')
        output = mse.forward(data=[torch_pred, torch_true], state={})
        self.assertTrue(np.allclose(output.detach().numpy(), 0.014375001))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import inspect
import random
from typing import Any, Dict, Iterable, List, Tuple, Union

import numpy as np
from PIL import Image, ImageOps

from fastestimator.op.numpyop.meta.one_of import OneOf
from fastestimator.op.numpyop.meta.sometimes import Sometimes
from fastestimator.op.numpyop.numpyop import NumpyOp, forward_numpyop
from fastestimator.op.numpyop.univariate.autocontrast import AutoContrast
from fastestimator.op.numpyop.univariate.brightness import Brightness
from fastestimator.op.numpyop.univariate.color import Color
from fastestimator.op.numpyop.univariate.contrast import Contrast
from fastestimator.op.numpyop.univariate.posterize import Posterize as PosterizeAug
from fastestimator.op.numpyop.univariate.sharpness import Sharpness
from fastestimator.op.numpyop.univariate.shear_x import ShearX
from fastestimator.op.numpyop.univariate.shear_y import ShearY
from fastestimator.op.numpyop.univariate.translate_x import TranslateX
from fastestimator.op.numpyop.univariate.translate_y import TranslateY
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import param_to_range, to_list, to_set


@traceable()
class Rotate(NumpyOp):
    """Rotate the input by an angle selected randomly.

    This is a wrapper for functionality provided by the PIL library:
    https://github.com/python-pillow/Pillow/tree/master/src/PIL.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        limit: Range from which the angle can be picked. If limit is a single int the range is considered from
            (0, limit).

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 limit: Union[int, Tuple[int, int]] = 30):
        super().__init__(inputs=to_list(inputs), outputs=to_list(outputs), mode=mode, ds_id=ds_id)
        self.limit = param_to_range(limit)

    def set_rua_level(self, magnitude_coef: float) -> None:
        """Set the augmentation intensity based on the magnitude_coef.

        This method is specifically designed to be invoked by the RUA Op.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """
        param_mid = (self.limit[1] + self.limit[0]) / 2
        param_extent = magnitude_coef * ((self.limit[1] - self.limit[0]) / 2)
        self.limit = (param_mid - param_extent, param_mid + param_extent)

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        degree = random.uniform(self.limit[0], self.limit[1])
        return [Rotate._apply_rotate(elem, degree) for elem in data]

    @staticmethod
    def _apply_rotate(data: np.ndarray, degree: float) -> np.ndarray:
        """Rotate the image.

        Args:
            data: The image to be modified.
            degree: Angle for image rotation.

        Returns:
            The image after applying rotation.
        """
        im = Image.fromarray(data)
        im = im.rotate(degree)
        return np.array(im)


@traceable()
class Identity(NumpyOp):
    """Pass the input as-is.

    Args:
        inputs: Key(s) of images.
        outputs: Key(s) into which to write the images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        super().__init__(inputs=to_list(inputs), outputs=to_list(outputs), mode=mode, ds_id=ds_id)

    def set_rua_level(self, magnitude_coef: float) -> None:
        """A method which will be invoked by the RUA Op to adjust the augmentation intensity.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """


@traceable()
class Equalize(NumpyOp):
    """Equalize the image histogram.

    This is a wrapper for functionality provided by the PIL library:
    https://github.com/python-pillow/Pillow/tree/master/src/PIL.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None):
        super().__init__(inputs=to_list(inputs), outputs=to_list(outputs), mode=mode, ds_id=ds_id)

    def set_rua_level(self, magnitude_coef: float) -> None:
        """A method which will be invoked by the RUA Op to adjust the augmentation intensity.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        return [Equalize._apply_equalize(elem) for elem in data]

    @staticmethod
    def _apply_equalize(data: np.ndarray) -> np.ndarray:
        """Equalize the image histogram.

        Args:
            data: The image to be modified.

        Returns:
            The image after applying equalize.
        """
        im = Image.fromarray(data)
        im = ImageOps.equalize(im)
        return np.array(im)


@traceable()
class Posterize(PosterizeAug):
    """Reduce the number of bits for the image.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        num_bits: Number of high bits. If num_bits is a single value, the range will be [num_bits, num_bits]. A triplet
            of ints will be interpreted as [r, g, b], and a triplet of pairs as [[r1, r1], [g1, g2], [b1, b2]]. Must be
            in the range [0, 8].

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 num_bits: Union[int,
                                 Tuple[int, int],
                                 Tuple[int, int, int],
                                 Tuple[Tuple[int, int], Tuple[int, int], Tuple[int, int]]] = 7):
        self.num_bits = num_bits
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id, num_bits=num_bits)

    def set_rua_level(self, magnitude_coef: float) -> None:
        """Set the augmentation intensity based on the magnitude_coef.

        This method is specifically designed to be invoked by the RUA Op.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """
        if isinstance(self.num_bits, tuple) and len(self.num_bits) == 3:
            num_bits = []
            for i in self.num_bits:
                num_bits.append(Posterize._range_tuple(num_bits=i, magnitude_coef=magnitude_coef))
            self.num_bits = tuple(num_bits)
        else:
            self.num_bits = Posterize._range_tuple(num_bits=self.num_bits, magnitude_coef=magnitude_coef)
        super().__init__(inputs=self.inputs,
                         outputs=self.outputs,
                         mode=self.mode,
                         ds_id=self.ds_id,
                         num_bits=self.num_bits)

    @staticmethod
    def _range_tuple(num_bits: Union[int, Tuple[int, int]], magnitude_coef: float) -> Tuple[int, int]:
        """Process num_bits for posterization based on augmentation intensity.

        Args:
            num_bits: Number of high bits.
            magnitude_coef: The desired augmentation intensity (range [0-1]).

        Returns:
            The range of high bits after adjusting augmentation intensity.
        """
        if isinstance(num_bits, tuple):
            param_mid = (num_bits[0] + num_bits[1])/2
            param_extent = magnitude_coef * ((num_bits[1] - num_bits[0])/2)
            bits_range = (round(param_mid - param_extent), round(param_mid + param_extent))
        else:
            bits_range = (round(8-(magnitude_coef*num_bits)), 8)
        return bits_range


@traceable()
class Solarize(NumpyOp):
    """Invert all pixel values above a threshold.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        threshold: Range for the solarizing threshold. If threshold is a single value 't', the range will be [0, t].

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 threshold: Union[int, Tuple[int, int], float, Tuple[float, float]] = 256):
        super().__init__(inputs=to_list(inputs), outputs=to_list(outputs), mode=mode, ds_id=ds_id)
        self.threshold = threshold

    def set_rua_level(self, magnitude_coef: Union[int, float]) -> None:
        """Set the augmentation intensity based on the magnitude_coef.

        This method is specifically designed to be invoked by the RUA Op.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
        """
        if isinstance(self.threshold, tuple):
            self.threshold = magnitude_coef * (self.threshold[1] - self.threshold[0]) + self.threshold[0]
        else:
            self.threshold = magnitude_coef * self.threshold

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        if isinstance(self.threshold, tuple):
            threshold = 256 - round(random.uniform(self.threshold[0], self.threshold[1]))
        else:
            threshold = 256 - round(random.uniform(0, self.threshold))
        return [Solarize._apply_solarize(elem, threshold) for elem in data]

    @staticmethod
    def _apply_solarize(data: np.ndarray, threshold: int) -> np.ndarray:
        """Invert all pixel values of the image above a threshold.

        Args:
            data: The image to be modified.
            threshold: Solarizing threshold.

        Returns:
            The image after applying solarize.
        """
        data = np.where(data < threshold, data, 255 - data)
        return data


@traceable()
class OneOfMultiVar(OneOf):
    """Perform one of several possible NumpyOps.

    Note that OneOfMultiVar accepts both univariate and multivariate ops and allows the list of passed NumpyOps to have
    different input and output keys. OneOfMultiVar should not be used to wrap an op whose output key(s) do not already
    exist in the data dictionary. This would result in a problem when future ops / traces attempt to reference the
    output key, but OneOfMultiVar declined to generate it. If you want to create a default value for a new key, simply
    use a LambdaOp before invoking the OneOfMultiVar.

    Args:
        *numpy_ops: A list of ops to choose between with uniform probability.
    """
    def __init__(self, *numpy_ops: NumpyOp) -> None:
        inputs = to_set(numpy_ops[0].inputs)
        outputs = to_set(numpy_ops[0].outputs)
        mode = numpy_ops[0].mode
        ds_id = numpy_ops[0].ds_id
        self.in_list = numpy_ops[0].in_list
        self.out_list = numpy_ops[0].out_list
        for op in numpy_ops[1:]:
            assert self.in_list == op.in_list, "All ops within OneOf must share the same input configuration"
            assert self.out_list == op.out_list, "All ops within OneOf must share the same output configuration"
            assert mode == op.mode, "All ops within a OneOf must share the same mode"

            for inp in op.inputs:
                inputs.add(inp)

            for out in op.outputs:
                outputs.add(out)

        # Bypassing OneOf Op's restriction of same input and output key(s) on the list of passed NumpyOps.
        super(OneOf, self).__init__(inputs=inputs.union(outputs), outputs=outputs, mode=mode, ds_id=ds_id)
        self.ops = numpy_ops

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        data = {key: elem for key, elem in zip(self.inputs, data)}
        forward_numpyop([random.choice(self.ops)], data, state)
        return [data[key] for key in self.outputs]


@traceable()
class RUA(NumpyOp):
    """Apply RUA augmentation strategy.

    Note that all augmentation ops passed to RUA should have a set_rua_level method to modify their strength based on
    the level. Custom NumpyOps can be passed to the `choices` argument along with names of augmentations to add. Passing
    'defaults' adds the default list of augmentations along with any custom NumpyOps specified by the user.
    The default augmentations are: 'Rotate', 'Identity', 'AutoContrast', 'Equalize', 'Posterize', 'Solarize',
    'Sharpness', 'Contrast', 'Color', 'Brightness', 'ShearX', 'ShearY', 'TranslateX' and 'TranslateY'.
    To add specific augmentations from the default list, their names can be passed. Ex: 'Rotate'.
    To remove specific augmentations from the list, you can negate their names. Ex: '!Rotate' will load all the
    augmentations except 'Rotate'.

    Example combinations which are not allowed:
    choices = ['defaults', 'Rotate']        # augmentations from the default list are redundant with 'defaults'.
    choices = ['defaults', '!Rotate']       # negated augmentations automatically load the default list.
    choices = ['!Solarize', 'Rotate']       # Cannot mix negated and normal augmentations.

    RUA should not have augmentation ops whose output key(s) do not already exist in the data dictionary. This would
    result in a problem when future ops / traces attempt to reference the output key, but RUA declined to generate it.
    If you want to create a default value for a new key, simply use a LambdaOp before invoking RUA.

    Args:
        inputs: Key(s) of images to be modified.
        outputs: Key(s) into which to write the modified images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        choices: List of augmentations to apply.
        level: Factor to set the range for magnitude of augmentation. Must be in the range [0, 30].

    Image types:
        uint8
    """
    def __init__(self,
                 inputs: Union[str, Iterable[str]],
                 outputs: Union[str, Iterable[str]],
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 choices: Union[str, NumpyOp, List[Union[str, NumpyOp]]] = "defaults",
                 level: Union[int, float] = 18):
        self.default_aug_dict = {
            "Rotate": Rotate(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,limit=90),
            "Identity": Identity(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id),
            "AutoContrast": AutoContrast(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id),
            "Equalize": Equalize(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id),
            "Posterize": Posterize(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,num_bits=7),
            "Solarize": Solarize(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,threshold=256),
            "Sharpness": Sharpness(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,limit=0.9),
            "Contrast": Contrast(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,limit=0.9),
            "Color": Color(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,limit=0.9),
            "Brightness": Brightness(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,limit=0.9),
            "ShearX": ShearX(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,shear_coef=0.5),
            "ShearY": ShearY(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,shear_coef=0.5),
            "TranslateX": TranslateX(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,shift_limit=0.33),
            "TranslateY": TranslateY(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id,shift_limit=0.33)
        }
        aug_options = self._parse_aug_choices(magnitude_coef=(level / 30.), choices=to_list(choices))

        inputs, outputs = to_set(inputs), to_set(outputs)
        for op in aug_options:
            for inp in op.inputs:
                inputs.add(inp)

            for out in op.outputs:
                outputs.add(out)
        super().__init__(inputs=inputs.union(outputs), outputs=outputs, mode=mode, ds_id=ds_id)

        # Calculating number of augmentation to apply at each training iteration
        N_min = 1
        N_max = min(len(aug_options), 5)
        N = level * (N_max - N_min) / 30 + N_min
        N_guarantee, N_p = int(N), N % 1

        self.ops = [OneOfMultiVar(*aug_options) for _ in range(N_guarantee)]
        if N_p > 0:
            self.ops.append(Sometimes(OneOfMultiVar(*aug_options), prob=N_p))

    def _parse_aug_choices(self, magnitude_coef: float, choices: List[Union[str, NumpyOp]]) -> List[NumpyOp]:
        """Parse the augmentation choices to determine the final list of augmentations to apply.

        Args:
            magnitude_coef: The desired augmentation intensity (range [0-1]).
            choices: List of augmentations to apply.

        Returns:
            List of augmentations to apply.

        Raises:
            AssertionError: If augmentations to add and remove are mixed.
            AttributeError: If augmentation choices don't have a 'set_rua_level' method.
            ValueError: If 'defaults' is provided with augmentation strings to add or remove, or wrong names are
                provided.
        """
        custom_ops = [op for op in choices if not isinstance(op, str)]
        remove_ops = [op for op in choices if isinstance(op, str) and op.startswith("!")]
        add_ops = [op for op in choices if isinstance(op, str) and not (op.startswith("!") or (op == "defaults"))]
        aug_names = list(self.default_aug_dict.keys())

        assert len(remove_ops)==0 or len(add_ops)==0, \
            "RUA supports either add or remove ops, but not both. Found {} and {}".format(add_ops, remove_ops)

        if len(remove_ops) > 0:
            if "defaults" in choices:
                raise ValueError("Can't provide 'defaults' value with ops to remove, found: {}".format(remove_ops))
            remove_ops = [op[1:] for op in remove_ops]

            for op in remove_ops:
                if op not in aug_names:
                    raise ValueError("Unable to remove {}, list of augmentations available: {}".format(op, aug_names))

            aug_list = [aug for aug_name, aug in self.default_aug_dict.items() if aug_name not in remove_ops]
        else:
            if "defaults" in choices:
                if len(add_ops) > 0:
                    raise ValueError("Can't pass 'defaults' value with default list's ops, found: {}".format(add_ops))
                aug_list = list(self.default_aug_dict.values())
            elif len(add_ops) > 0:
                for op in add_ops:
                    if op not in aug_names:
                        raise ValueError("Unable to add {}, list of augmentations available: {}".format(op, aug_names))

                aug_list = [self.default_aug_dict[aug_name] for aug_name in add_ops]
            else:
                aug_list = []
        aug_list = aug_list + custom_ops

        for op in aug_list:
            if hasattr(op, "set_rua_level") and inspect.ismethod(getattr(op, "set_rua_level")):
                op.set_rua_level(magnitude_coef=magnitude_coef)
            else:
                raise AttributeError(
                    "RUA Augmentations should have a 'set_rua_level' method but it's not present in Op: {}".format(
                        op.__class__.__name__))

        return aug_list

    def forward(self, data: List[np.ndarray], state: Dict[str, Any]) -> List[np.ndarray]:
        data = {key: elem for key, elem in zip(self.inputs, data)}
        forward_numpyop(self.ops, data, state)
        return [data[key] for key in self.outputs]



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import contextlib
import locale
import os
import platform
import re
import shutil
import sys
import types
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Set, Tuple, Union, Optional
from unittest.mock import Base, MagicMock

import dot2tex as d2t
import jsonpickle
import matplotlib
import numpy as np
import pydot
import pytorch_model_summary as pms
import tensorflow as tf
import torch
from natsort import humansorted
from pylatex import Command, Document, Figure, Hyperref, Itemize, Label, LongTable, Marker, MultiColumn, NoEscape, \
    Package, Section, Subsection, Subsubsection, Tabularx, escape_latex
from pylatex.base_classes import Arguments
from pylatex.section import Paragraph
from pylatex.utils import bold
from torch.utils.data import Dataset

import fastestimator as fe
from fastestimator.dataset.dataset import FEDataset
from fastestimator.network import BaseNetwork
from fastestimator.op.numpyop.meta.fuse import Fuse
from fastestimator.op.numpyop.meta.one_of import OneOf
from fastestimator.op.numpyop.meta.repeat import Repeat
from fastestimator.op.numpyop.meta.sometimes import Sometimes
from fastestimator.op.op import Op
from fastestimator.op.tensorop.meta.fuse import Fuse as FuseT
from fastestimator.op.tensorop.meta.one_of import OneOf as OneOfT
from fastestimator.op.tensorop.meta.repeat import Repeat as RepeatT
from fastestimator.op.tensorop.meta.sometimes import Sometimes as SometimesT
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.pipeline import Pipeline
from fastestimator.schedule.schedule import Scheduler, get_current_items, get_signature_epochs
from fastestimator.summary.logs.log_plot import visualize_logs
from fastestimator.trace.io.restore_wizard import RestoreWizard
from fastestimator.trace.trace import Trace, sort_traces
from fastestimator.util.data import Data
from fastestimator.util.latex_util import AdjustBox, Center, ContainerList, HrefFEID, Verbatim
from fastestimator.util.traceability_util import FeSummaryTable, traceable
from fastestimator.util.util import FEID, LogSplicer, Suppressor, prettify_metric_name, to_list, NonContext


@traceable()
class Traceability(Trace):
    """Automatically generate summary reports of the training.

    Args:
        save_path: Where to save the output files. Note that this will generate a new folder with the given name, into
            which the report and corresponding graphics assets will be written.
        extra_objects: Any extra objects which are not part of the Estimator, but which you want to capture in the
            summary report. One example could be an extra pipeline which performs pre-processing.

    Raises:
        OSError: If graphviz is not installed.
    """
    def __init__(self, save_path: str, extra_objects: Any = None):
        # Verify that graphviz is available on this machine
        try:
            pydot.Dot.create(pydot.Dot())
        except OSError:
            raise OSError(
                "Traceability requires that graphviz be installed. See www.graphviz.org/download for more information.")
        # Verify that the system locale is functioning correctly
        try:
            locale.getlocale()
        except ValueError:
            raise OSError("Your system locale is not configured correctly. On mac this can be resolved by adding \
                'export LC_ALL=en_US.UTF-8' and 'export LANG=en_US.UTF-8' to your ~/.bash_profile")
        super().__init__(inputs="*", mode="!infer")  # Claim wildcard inputs to get this trace sorted last
        # Report assets will get saved into a folder for portability
        path = os.path.normpath(save_path)
        path = os.path.abspath(path)
        root_dir = os.path.dirname(path)
        report = os.path.basename(path) or 'report'
        report = report.split('.')[0]
        self.save_dir = os.path.join(root_dir, report)
        self.resource_dir = os.path.join(self.save_dir, 'resources')
        self.report_name = None  # This will be set later by the experiment name
        os.makedirs(self.save_dir, exist_ok=True)
        os.makedirs(self.resource_dir, exist_ok=True)
        # Other member variables
        self.config_tables = []
        # Extra objects will automatically get included in the report since this Trace is @traceable, so we don't need
        # to do anything with them. Referencing here to stop IDEs from flagging the argument as unused and removing it.
        to_list(extra_objects)
        self.doc = Document()
        self.log_splicer = None

    def on_begin(self, data: Data) -> None:
        exp_name = self.system.summary.name
        if not exp_name:
            raise RuntimeError("Traceability reports require an experiment name to be provided in estimator.fit()")
        # Convert the experiment name to a report name (useful for saving multiple experiments into same directory)
        report_name = "".join('_' if c == ' ' else c for c in exp_name
                              if c.isalnum() or c in (' ', '_')).rstrip().lower()
        report_name = re.sub('_{2,}', '_', report_name)
        self.report_name = report_name or 'report'
        # Send experiment logs into a file
        log_path = os.path.join(self.resource_dir, f"{report_name}.txt")
        if self.system.mode != 'test':
            # See if there's a RestoreWizard
            restore = False
            for trace in self.system.traces:
                if isinstance(trace, RestoreWizard):
                    restore = trace.should_restore()
            if not restore:
                # If not running in test mode, we need to remove any old log file since it would get appended to
                with contextlib.suppress(FileNotFoundError):
                    os.remove(log_path)
        self.log_splicer = LogSplicer(log_path)
        self.log_splicer.__enter__()
        # Get the initialization summary information for the experiment
        self.config_tables = self.system.summary.system_config
        models = self.system.network.models
        n_floats = len(self.config_tables) + len(models)

        self.doc = self._init_document_geometry()
        # Keep tables/figures in their sections
        self.doc.packages.append(Package(name='placeins', options=['section']))
        self.doc.preamble.append(NoEscape(r'\usetikzlibrary{positioning}'))

        # Fix an issue with too many tables for LaTeX to render
        self.doc.preamble.append(NoEscape(r'\maxdeadcycles=' + str(2 * n_floats + 10) + ''))
        self.doc.preamble.append(NoEscape(r'\extrafloats{' + str(n_floats + 10) + '}'))

        # Manipulate booktab tables so that their horizontal lines don't break
        self.doc.preamble.append(NoEscape(r'\aboverulesep=0ex'))
        self.doc.preamble.append(NoEscape(r'\belowrulesep=0ex'))
        self.doc.preamble.append(NoEscape(r'\renewcommand{\arraystretch}{1.2}'))

        self._write_title()
        self._write_toc()

    def on_end(self, data: Data) -> None:
        self._write_body_content()

        # Need to move the tikz dependency after the xcolor package
        self.doc.dumps_packages()
        packages = self.doc.packages
        tikz = Package(name='tikz')
        packages.discard(tikz)
        packages.add(tikz)

        if shutil.which("latexmk") is None and shutil.which("pdflatex") is None:
            # No LaTeX Compiler is available
            self.doc.generate_tex(os.path.join(self.save_dir, self.report_name))
            suffix = '.tex'
        else:
            # Force a double-compile since some compilers will struggle with TOC generation
            self.doc.generate_pdf(os.path.join(self.save_dir, self.report_name), clean_tex=False, clean=False)
            self.doc.generate_pdf(os.path.join(self.save_dir, self.report_name), clean_tex=False)
            suffix = '.pdf'
        print("FastEstimator-Traceability: Report written to {}{}".format(os.path.join(self.save_dir, self.report_name),
                                                                          suffix))
        self.log_splicer.__exit__()

    def _write_title(self) -> None:
        """Write the title content of the file. Override if you want to build on top of base traceability report.
        """
        self.doc.preamble.append(Command('title', self.system.summary.name))
        self.doc.preamble.append(Command('author', f"FastEstimator {fe.__version__}"))
        self.doc.preamble.append(Command('date', NoEscape(r'\today')))
        self.doc.append(NoEscape(r'\maketitle'))

    def _write_toc(self) -> None:
        """Write the table of contents. Override if you want to build on top of base traceability report.
        """
        self.doc.append(NoEscape(r'\tableofcontents'))
        self.doc.append(NoEscape(r'\newpage'))

    def _write_body_content(self) -> None:
        """Write the main content of the file. Override if you want to build on top of base traceability report.
        """
        self._document_training_graphs()
        self.doc.append(NoEscape(r'\newpage'))
        self._document_fe_graph()
        self.doc.append(NoEscape(r'\newpage'))
        self._document_init_params()
        self._document_models()
        self._document_sys_config()
        self.doc.append(NoEscape(r'\newpage'))

    def _document_training_graphs(self) -> None:
        """Add training graphs to the traceability document.
        """
        with self.doc.create(Section("Training Graphs")):
            log_path = os.path.join(self.resource_dir, f'{self.report_name}_logs.png')
            visualize_logs(experiments=[self.system.summary],
                           save_path=log_path,
                           verbose=False,
                           ignore_metrics={'num_device', 'logging_interval'})
            with self.doc.create(Figure(position='h!')) as plot:
                plot.add_image(os.path.relpath(log_path, start=self.save_dir),
                               width=NoEscape(r'1.0\textwidth,height=0.95\textheight,keepaspectratio'))
            for idx, graph in enumerate(self.system.custom_graphs.values()):
                graph_path = os.path.join(self.resource_dir, f'{self.report_name}_custom_graph_{idx}.png')
                visualize_logs(experiments=graph, save_path=graph_path, verbose=False)
                with self.doc.create(Figure(position='h!')) as plot:
                    plot.add_image(os.path.relpath(graph_path, start=self.save_dir),
                                   width=NoEscape(r'1.0\textwidth,height=0.95\textheight,keepaspectratio'))

    def _document_fe_graph(self) -> None:
        """Add FE execution graphs into the traceability document.
        """
        with self.doc.create(Section("FastEstimator Architecture")):
            for mode in self.system.pipeline.data.keys():
                scheduled_items = self.system.pipeline.get_scheduled_items(
                    mode) + self.system.network.get_scheduled_items(mode) + self.system.traces
                signature_epochs = get_signature_epochs(scheduled_items, total_epochs=self.system.epoch_idx, mode=mode)
                epochs_with_data = self.system.pipeline.get_epochs_with_data(total_epochs=self.system.epoch_idx,
                                                                             mode=mode)
                if set(signature_epochs) & epochs_with_data:
                    self.doc.append(NoEscape(r'\FloatBarrier'))
                    with self.doc.create(Subsection(mode.capitalize())):
                        for epoch in signature_epochs:
                            if epoch not in epochs_with_data:
                                continue
                            self.doc.append(NoEscape(r'\FloatBarrier'))
                            with self.doc.create(
                                    Subsubsection(f"Epoch {epoch}",
                                                  label=Label(Marker(name=f"{mode}{epoch}", prefix="ssubsec")))):
                                ds_ids = self.system.pipeline.get_ds_ids(epoch=epoch, mode=mode)
                                for ds_id in ds_ids:
                                    with NonContext() if ds_id is None else self.doc.create(
                                            Paragraph(f"Dataset {ds_id}",
                                                      label=Label(Marker(name=f"{mode}{epoch}{ds_id}",
                                                                         prefix="para")))):
                                        diagram = self._draw_diagram(mode, epoch, ds_id)
                                        ltx = d2t.dot2tex(diagram.to_string(), figonly=True)
                                        args = Arguments(**{'max width': r'\textwidth, max height=0.9\textheight'})
                                        args.escape = False
                                        with self.doc.create(Center()):
                                            with self.doc.create(AdjustBox(arguments=args)) as box:
                                                box.append(NoEscape(ltx))

    def _document_init_params(self) -> None:
        """Add initialization parameters to the traceability document.
        """
        from fastestimator.estimator import Estimator  # Avoid circular import
        with self.doc.create(Section("Parameters")):
            model_ids = {
                FEID(id(model))
                for model in self.system.network.models if isinstance(model, (tf.keras.Model, torch.nn.Module))
            }
            # Locate the datasets in order to provide extra details about them later in the summary
            datasets = {}
            for mode in ['train', 'eval', 'test']:
                objs = to_list(self.system.pipeline.data.get(mode, None))
                idx = 0
                while idx < len(objs):
                    obj = objs[idx]
                    if obj:
                        feid = FEID(id(obj))
                        if feid not in datasets:
                            datasets[feid] = ({mode}, obj)
                        else:
                            datasets[feid][0].add(mode)
                    if isinstance(obj, Scheduler):
                        objs.extend(obj.get_all_values())
                    idx += 1
            # Parse the config tables
            start = 0
            start = self._loop_tables(start,
                                      classes=(Estimator, BaseNetwork, Pipeline),
                                      name="Base Classes",
                                      model_ids=model_ids,
                                      datasets=datasets)
            start = self._loop_tables(start,
                                      classes=Scheduler,
                                      name="Schedulers",
                                      model_ids=model_ids,
                                      datasets=datasets)
            start = self._loop_tables(start, classes=Trace, name="Traces", model_ids=model_ids, datasets=datasets)
            start = self._loop_tables(start, classes=Op, name="Operators", model_ids=model_ids, datasets=datasets)
            start = self._loop_tables(start,
                                      classes=(Dataset, tf.data.Dataset),
                                      name="Datasets",
                                      model_ids=model_ids,
                                      datasets=datasets)
            start = self._loop_tables(start,
                                      classes=(tf.keras.Model, torch.nn.Module),
                                      name="Models",
                                      model_ids=model_ids,
                                      datasets=datasets)
            start = self._loop_tables(start,
                                      classes=types.FunctionType,
                                      name="Functions",
                                      model_ids=model_ids,
                                      datasets=datasets)
            start = self._loop_tables(start,
                                      classes=(np.ndarray, tf.Tensor, tf.Variable, torch.Tensor),
                                      name="Tensors",
                                      model_ids=model_ids,
                                      datasets=datasets)
            self._loop_tables(start, classes=Any, name="Miscellaneous", model_ids=model_ids, datasets=datasets)

    def _loop_tables(self,
                     start: int,
                     classes: Union[type, Tuple[type, ...]],
                     name: str,
                     model_ids: Set[FEID],
                     datasets: Dict[FEID, Tuple[Set[str], Any]]) -> int:
        """Iterate through tables grouping them into subsections.

        Args:
            start: What index to start searching from.
            classes: What classes are acceptable for this subsection.
            name: What to call this subsection.
            model_ids: The ids of any known models.
            datasets: A mapping like {ID: ({modes}, dataset)}. Useful for augmenting the displayed information.

        Returns:
            The new start index after traversing as many spaces as possible along the list of tables.
        """
        stop = start
        while stop < len(self.config_tables):
            if classes == Any or issubclass(self.config_tables[stop].type, classes):
                stop += 1
            else:
                break
        if stop > start:
            self.doc.append(NoEscape(r'\FloatBarrier'))
            with self.doc.create(Subsection(name)):
                self._write_tables(self.config_tables[start:stop], model_ids, datasets)
        return stop

    def _write_tables(self,
                      tables: List[FeSummaryTable],
                      model_ids: Set[FEID],
                      datasets: Dict[FEID, Tuple[Set[str], Any]]) -> None:
        """Insert a LaTeX representation of a list of tables into the current doc.

        Args:
            tables: The tables to write into the doc.
            model_ids: The ids of any known models.
            datasets: A mapping like {ID: ({modes}, dataset)}. Useful for augmenting the displayed information.
        """
        for tbl in tables:
            name_override = None
            toc_ref = None
            extra_rows = None
            if tbl.fe_id in model_ids:
                # Link to a later detailed model description
                name_override = Hyperref(Marker(name=str(tbl.name), prefix="subsec"),
                                         text=NoEscape(r'\textcolor{blue}{') + bold(tbl.name) + NoEscape('}'))
            if tbl.fe_id in datasets:
                modes, dataset = datasets[tbl.fe_id]
                title = ", ".join([s.capitalize() for s in modes])
                name_override = bold(f'{tbl.name} ({title})')
                # Enhance the dataset summary
                if isinstance(dataset, FEDataset):
                    extra_rows = list(dataset.summary().__getstate__().items())
                    for idx, (key, val) in enumerate(extra_rows):
                        key = f"{prettify_metric_name(key)}:"
                        if isinstance(val, dict) and val:
                            if isinstance(list(val.values())[0], (int, float, str, bool, type(None))):
                                val = jsonpickle.dumps(val, unpicklable=False)
                            else:
                                subtable = Tabularx('l|X', width_argument=NoEscape(r'\linewidth'))
                                for k, v in val.items():
                                    if hasattr(v, '__getstate__'):
                                        v = jsonpickle.dumps(v, unpicklable=False)
                                    subtable.add_row((k, v))
                                # To nest TabularX, have to wrap it in brackets
                                subtable = ContainerList(data=[NoEscape("{"), subtable, NoEscape("}")])
                                val = subtable
                        extra_rows[idx] = (key, val)
            tbl.render_table(self.doc, name_override=name_override, toc_ref=toc_ref, extra_rows=extra_rows)

    def _document_models(self) -> None:
        """Add model summaries to the traceability document.
        """
        with self.doc.create(Section("Models")):
            for model in humansorted(self.system.network.models, key=lambda m: m.model_name):
                if not isinstance(model, (tf.keras.Model, torch.nn.Module)):
                    continue
                self.doc.append(NoEscape(r'\FloatBarrier'))
                with self.doc.create(Subsection(f"{model.model_name.capitalize()}")):
                    if isinstance(model, tf.keras.Model):
                        # Text Summary
                        summary = []
                        model.summary(line_length=92, print_fn=lambda x: summary.append(x))
                        summary = "\n".join(summary)
                        self.doc.append(Verbatim(summary))
                        with self.doc.create(Center()):
                            self.doc.append(HrefFEID(FEID(id(model)), model.model_name))

                        # Visual Summary
                        # noinspection PyBroadException
                        try:
                            file_path = os.path.join(self.resource_dir,
                                                     "{}_{}.pdf".format(self.report_name, model.model_name))
                            dot = tf.keras.utils.model_to_dot(model, show_shapes=True, expand_nested=True)
                            # LaTeX \maxdim is around 575cm (226 inches), so the image must have max dimension less than
                            # 226 inches. However, the 'size' parameter doesn't account for the whole node height, so
                            # set the limit lower (100 inches) to leave some wiggle room.
                            dot.set('size', '100')
                            dot.write(file_path, format='pdf')
                        except Exception:
                            file_path = None
                            print(
                                f"FastEstimator-Warn: Model {model.model_name} could not be visualized by Traceability")
                    elif isinstance(model, torch.nn.Module):
                        if hasattr(model, 'fe_input_spec'):
                            # Text Summary
                            # noinspection PyUnresolvedReferences
                            inputs = model.fe_input_spec.get_dummy_input()
                            self.doc.append(
                                Verbatim(
                                    pms.summary(model.module if self.system.num_devices > 1 else model,
                                                inputs,
                                                print_summary=False)))
                            with self.doc.create(Center()):
                                self.doc.append(HrefFEID(FEID(id(model)), model.model_name))
                            # Visual Summary
                            # Import has to be done while matplotlib is using the Agg backend
                            old_backend = matplotlib.get_backend() or 'Agg'
                            matplotlib.use('Agg')
                            # noinspection PyBroadException
                            try:
                                # Fake the IPython import when user isn't running from Jupyter
                                sys.modules.setdefault('IPython', MagicMock())
                                sys.modules.setdefault('IPython.display', MagicMock())
                                import hiddenlayer as hl
                                with Suppressor():
                                    graph = hl.build_graph(model.module if self.system.num_devices > 1 else model,
                                                           inputs)
                                graph = graph.build_dot()
                                graph.attr(rankdir='TB')  # Switch it to Top-to-Bottom instead of Left-to-Right
                                # LaTeX \maxdim is around 575cm (226 inches), so the image must have max dimension less
                                # than 226 inches. However, the 'size' parameter doesn't account for the whole node
                                # height, so set the limit lower (100 inches) to leave some wiggle room.
                                graph.attr(size="100,100")
                                graph.attr(margin='0')
                                file_path = graph.render(filename="{}_{}".format(self.report_name, model.model_name),
                                                         directory=self.resource_dir,
                                                         format='pdf',
                                                         cleanup=True)
                            except Exception:
                                file_path = None
                                print("FastEstimator-Warn: Model {} could not be visualized by Traceability".format(
                                    model.model_name))
                            finally:
                                matplotlib.use(old_backend)
                        else:
                            file_path = None
                            self.doc.append("This model was not used by the Network during training.")
                    if file_path:
                        with self.doc.create(Figure(position='ht!')) as fig:
                            fig.append(Label(Marker(name=str(FEID(id(model))), prefix="model")))
                            fig.add_image(os.path.relpath(file_path, start=self.save_dir),
                                          width=NoEscape(r'1.0\textwidth,height=0.95\textheight,keepaspectratio'))
                            fig.add_caption(NoEscape(HrefFEID(FEID(id(model)), model.model_name).dumps()))

    def _document_sys_config(self) -> None:
        """Add a system config summary to the traceability document.
        """
        with self.doc.create(Section("System Configuration")):
            with self.doc.create(Itemize()) as itemize:
                itemize.add_item(escape_latex(f"FastEstimator {fe.__version__}"))
                itemize.add_item(escape_latex(f"Python {platform.python_version()}"))
                itemize.add_item(escape_latex(f"OS: {sys.platform}"))
                itemize.add_item(f"Number of GPUs: {torch.cuda.device_count()}")
                if fe.fe_deterministic_seed is not None:
                    itemize.add_item(escape_latex(f"Deterministic Seed: {fe.fe_deterministic_seed}"))
            with self.doc.create(LongTable('|lr|', pos=['h!'], booktabs=True)) as tabular:
                tabular.add_row((bold("Module"), bold("Version")))
                tabular.add_hline()
                tabular.end_table_header()
                tabular.add_hline()
                tabular.add_row((MultiColumn(2, align='r', data='Continued on Next Page'), ))
                tabular.add_hline()
                tabular.end_table_footer()
                tabular.end_table_last_footer()
                color = True
                for name, module in humansorted(sys.modules.items(), key=lambda x: x[0]):
                    if "." in name:
                        continue  # Skip sub-packages
                    if name.startswith("_"):
                        continue  # Skip private packages
                    if isinstance(module, Base):
                        continue  # Skip fake packages we mocked
                    if hasattr(module, '__version__'):
                        tabular.add_row((escape_latex(name), escape_latex(str(module.__version__))),
                                        color='black!5' if color else 'white')
                        color = not color
                    elif hasattr(module, 'VERSION'):
                        tabular.add_row((escape_latex(name), escape_latex(str(module.VERSION))),
                                        color='black!5' if color else 'white')
                        color = not color

    def _draw_diagram(self, mode: str, epoch: int, ds_id: Optional[str]) -> pydot.Dot:
        """Draw a summary diagram of the FastEstimator Ops / Traces.

        Args:
            mode: The execution mode to summarize ('train', 'eval', 'test', or 'infer').
            epoch: The epoch to summarize.
            ds_id: The ds_id to summarize.

        Returns:
            A pydot digraph representing the execution flow.
        """
        ds = self.system.pipeline.data[mode][ds_id]
        if isinstance(ds, Scheduler):
            ds = ds.get_current_value(epoch)
        pipe_ops = get_current_items(self.system.pipeline.ops, run_modes=mode, epoch=epoch, ds_id=ds_id) if isinstance(
            ds, Dataset) else []
        net_ops = get_current_items(self.system.network.ops, run_modes=mode, epoch=epoch, ds_id=ds_id)
        net_post = get_current_items(self.system.network.postprocessing, run_modes=mode, epoch=epoch, ds_id=ds_id)
        traces = sort_traces(get_current_items(self.system.traces, run_modes=mode, epoch=epoch, ds_id=ds_id),
                             ds_ids=self.system.pipeline.get_ds_ids(epoch=epoch, mode=mode))
        diagram = pydot.Dot(compound='true')  # Compound lets you draw edges which terminate at sub-graphs
        diagram.set('rankdir', 'TB')
        diagram.set('dpi', 300)
        diagram.set_node_defaults(shape='box')

        # Make the dataset the first of the pipeline ops
        pipe_ops.insert(0, ds)
        label_last_seen = defaultdict(lambda: str(id(ds)))  # Where was this key last generated

        batch_size = ""
        if isinstance(ds, Dataset):
            if hasattr(ds, "fe_batch") and ds.fe_batch:
                batch_size = ds.fe_batch
            else:
                batch_size = self.system.pipeline.batch_size
                if isinstance(batch_size, Scheduler):
                    batch_size = batch_size.get_current_value(epoch)
                if isinstance(batch_size, dict):
                    batch_size = batch_size[mode]
        if batch_size is not None:
            batch_size = f" (Batch Size: {batch_size})"
        self._draw_subgraph(diagram, diagram, label_last_seen, f'Pipeline{batch_size}', pipe_ops, ds_id)
        self._draw_subgraph(diagram, diagram, label_last_seen, 'Network', net_ops + net_post, ds_id)
        self._draw_subgraph(diagram, diagram, label_last_seen, 'Traces', traces, ds_id)
        return diagram

    def _draw_subgraph(self,
                       progenitor: pydot.Dot,
                       diagram: Union[pydot.Dot, pydot.Cluster],
                       label_last_seen: DefaultDict[str, str],
                       subgraph_name: str,
                       subgraph_ops: List[Union[Op, Trace, Any]],
                       ds_id: Optional[str]) -> None:
        """Draw a subgraph of ops into an existing `diagram`.

        Args:
            progenitor: The very top level diagram onto which Edges should be written.
            diagram: The diagram into which to add new Nodes.
            label_last_seen: A mapping of {data_dict_key: node_id} indicating the last node which generated the key.
            subgraph_name: The name to be associated with this subgraph.
            subgraph_ops: The ops to be wrapped in this subgraph.
            ds_id: The ds_id to be associated with this subgraph.
        """
        subgraph = pydot.Cluster(style='dashed', graph_name=subgraph_name, color='black')
        subgraph.set('label', subgraph_name)
        subgraph.set('labeljust', 'l')
        for idx, op in enumerate(subgraph_ops):
            node_id = str(id(op))
            self._add_node(progenitor, subgraph, op, label_last_seen, ds_id)
            if isinstance(op, Trace) and idx > 0:
                # Invisibly connect traces in order so that they aren't all just squashed horizontally into the image
                progenitor.add_edge(pydot.Edge(src=str(id(subgraph_ops[idx - 1])), dst=node_id, style='invis'))
        diagram.add_subgraph(subgraph)

    def _add_node(self,
                  progenitor: pydot.Dot,
                  diagram: Union[pydot.Dot, pydot.Cluster],
                  op: Union[Op, Trace, Any],
                  label_last_seen: DefaultDict[str, str],
                  ds_id: Optional[str],
                  edges: bool = True) -> None:
        """Draw a node onto a diagram based on a given op.

        Args:
            progenitor: The very top level diagram onto which Edges should be written.
            diagram: The diagram to be appended to.
            op: The op (or trace) to be visualized.
            label_last_seen: A mapping of {data_dict_key: node_id} indicating the last node which generated the key.
            ds_id: The ds_id under which the node is currently running.
            edges: Whether to write Edges to/from this Node.
        """
        node_id = str(id(op))
        if isinstance(op, (Sometimes, SometimesT)) and op.op:
            wrapper = pydot.Cluster(style='dotted', color='red', graph_name=str(id(op)))
            wrapper.set('label', f'Sometimes ({op.prob}):')
            wrapper.set('labeljust', 'l')
            edge_srcs = defaultdict(lambda: [])
            if op.extra_inputs:
                for inp in op.extra_inputs:
                    if inp == '*':
                        continue
                    edge_srcs[label_last_seen[inp]].append(inp)
            self._add_node(progenitor, wrapper, op.op, label_last_seen, ds_id)
            diagram.add_subgraph(wrapper)
            dst_id = self._get_all_nodes(wrapper)[0].get_name()
            for src, labels in edge_srcs.items():
                progenitor.add_edge(
                    pydot.Edge(src=src, dst=dst_id, lhead=wrapper.get_name(), label=f" {', '.join(labels)} "))
        elif isinstance(op, (OneOf, OneOfT)) and op.ops:
            wrapper = pydot.Cluster(style='dotted', color='darkorchid4', graph_name=str(id(op)))
            wrapper.set('label', 'One Of:')
            wrapper.set('labeljust', 'l')
            self._add_node(progenitor, wrapper, op.ops[0], label_last_seen, ds_id, edges=True)
            for sub_op in op.ops[1:]:
                self._add_node(progenitor, wrapper, sub_op, label_last_seen, ds_id, edges=False)
            diagram.add_subgraph(wrapper)
        elif isinstance(op, (Fuse, FuseT)) and op.ops:
            self._draw_subgraph(progenitor, diagram, label_last_seen, 'Fuse:', op.ops, ds_id)
        elif isinstance(op, (Repeat, RepeatT)) and op.op:
            wrapper = pydot.Cluster(style='dotted', color='darkgreen', graph_name=str(id(op)))
            wrapper.set('label', f'Repeat:')
            wrapper.set('labeljust', 'l')
            wrapper.add_node(
                pydot.Node(node_id,
                           label=f'{op.repeat if isinstance(op.repeat, int) else "?"}',
                           shape='doublecircle',
                           width=0.1))
            # dot2tex doesn't seem to handle edge color conversion correctly, so have to set hex color
            progenitor.add_edge(pydot.Edge(src=node_id + ":ne", dst=node_id + ":w", color='#006300'))
            self._add_node(progenitor, wrapper, op.op, label_last_seen, ds_id)
            # Add repeat edges
            edge_srcs = defaultdict(lambda: [])
            for out in op.outputs:
                if out in op.inputs and out not in op.repeat_inputs:
                    edge_srcs[label_last_seen[out]].append(out)
            for inp in op.repeat_inputs:
                edge_srcs[label_last_seen[inp]].append(inp)
            for src, labels in edge_srcs.items():
                progenitor.add_edge(pydot.Edge(src=src, dst=node_id, constraint=False, label=f" {', '.join(labels)} "))
            diagram.add_subgraph(wrapper)
        else:
            if isinstance(op, ModelOp):
                label = f"{op.__class__.__name__} ({FEID(id(op))}): {op.model.model_name}"
                model_ref = Hyperref(Marker(name=str(op.model.model_name), prefix='subsec'),
                                     text=NoEscape(r'\textcolor{blue}{') + bold(op.model.model_name) +
                                     NoEscape('}')).dumps()
                texlbl = f"{HrefFEID(FEID(id(op)), name=op.__class__.__name__).dumps()}: {model_ref}"
            else:
                label = f"{op.__class__.__name__} ({FEID(id(op))})"
                texlbl = HrefFEID(FEID(id(op)), name=op.__class__.__name__).dumps()
            diagram.add_node(pydot.Node(node_id, label=label, texlbl=texlbl))
            if isinstance(op, (Op, Trace)) and edges:
                # Need the instance check since subgraph_ops might contain a tf dataset or torch dataloader
                self._add_edge(progenitor, op, label_last_seen, ds_id)

    def _add_edge(self,
                  progenitor: pydot.Dot,
                  op: Union[Trace, Op],
                  label_last_seen: Dict[str, str],
                  ds_id: Optional[str]):
        """Draw edges into a given Node.

        Args:
            progenitor: The very top level diagram onto which Edges should be written.
            op: The op (or trace) to be visualized.
            label_last_seen: A mapping of {data_dict_key: node_id} indicating the last node which generated the key.
            ds_id: The ds_id under which the node is currently running.
        """
        node_id = str(id(op))
        edge_srcs = defaultdict(lambda: [])
        global_ds_ids = {key for vals in self.system.pipeline.data.values() for key in vals.keys() if key is not None}
        for inp in op.inputs:
            if inp == '*':
                continue
            _, candidate_id, *_ = f"{inp}|".split('|')
            if candidate_id in global_ds_ids and candidate_id != ds_id:
                continue  # Skip inputs which will be provided in other ds_id plots
            edge_srcs[label_last_seen[inp]].append(inp)
        for src, labels in edge_srcs.items():
            progenitor.add_edge(pydot.Edge(src=src, dst=node_id, label=f" {', '.join(labels)} "))
        outputs = op.get_outputs(ds_ids=ds_id) if isinstance(op, Trace) else op.outputs
        for out in outputs:
            label_last_seen[out] = node_id

    @staticmethod
    def _get_all_nodes(diagram: Union[pydot.Dot, pydot.Cluster]) -> List[pydot.Node]:
        """Recursively search through a `diagram` looking for Nodes.

        Args:
            diagram: The diagram to be inspected.

        Returns:
            All of the Nodes available within this diagram and its child diagrams.
        """
        nodes = diagram.get_nodes()
        for subgraph in diagram.get_subgraphs():
            nodes.extend(Traceability._get_all_nodes(subgraph))
        return nodes

    @staticmethod
    def _init_document_geometry() -> Document:
        """Init geometry setting of the document.

        Return:
            Initialized Document object.
        """
        return Document(geometry_options=['lmargin=2cm', 'rmargin=2cm', 'bmargin=2cm'])



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import multiprocessing
import os
import sqlite3 as sql
import sys
import traceback
from collections import defaultdict
from contextlib import closing
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Type

import tensorflow.keras.mixed_precision as mixed_precision
import torch
from prettytable import PrettyTable, from_db_cursor

from fastestimator.schedule.schedule import Scheduler
from fastestimator.summary.logs.log_parse import parse_log_iter
from fastestimator.summary.logs.log_plot import visualize_logs
from fastestimator.summary.summary import Summary, average_summaries
from fastestimator.summary.system import System
from fastestimator.util.cli_util import SaveAction
from fastestimator.util.util import NonContext, parse_string_to_python

_MAKE_HIST_TABLE = 'CREATE TABLE IF NOT EXISTS history (' \
                   'file TEXT, ' \
                   'experiment TEXT, ' \
                   'status TEXT, ' \
                   'args LIST[STR], ' \
                   'fe_version TEXT, ' \
                   'train_start TIMESTAMP, ' \
                   'train_end TIMESTAMP, ' \
                   'n_gpus INTEGER, ' \
                   'n_cpus INTEGER, ' \
                   'n_workers INTEGER, ' \
                   'n_restarts INTEGER, ' \
                   'pk INTEGER PRIMARY KEY' \
                   ')'
_MAKE_FEAT_TABLE = 'CREATE TABLE IF NOT EXISTS features (' \
                   'feature TEXT, ' \
                   'fk INTEGER REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                   ')'
_MAKE_DS_TABLE = 'CREATE TABLE IF NOT EXISTS datasets (' \
                 'mode TEXT, ' \
                 'dataset TEXT, ' \
                 'fk INTEGER REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                 ')'
_MAKE_PIPELINE_TABLE = 'CREATE TABLE IF NOT EXISTS pipeline (' \
                       'pipe_op TEXT, ' \
                       'fk INTEGER REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                       ')'
_MAKE_NETWORK_TABLE = 'CREATE TABLE IF NOT EXISTS network (' \
                      'net_op TEXT, ' \
                      'fk INTEGER REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                      ')'
_MAKE_POST_PROCESS_TABLE = 'CREATE TABLE IF NOT EXISTS postprocess (' \
                           'pp_op TEXT, ' \
                           'fk INTEGER REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                           ')'
_MAKE_TRACE_TABLE = 'CREATE TABLE IF NOT EXISTS traces (' \
                    'trace TEXT, ' \
                    'fk INTEGER REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                    ')'
_MAKE_ERR_TABLE = 'CREATE TABLE IF NOT EXISTS errors (' \
                  'exc_type TEXT, ' \
                  'exc_tb TEXT, ' \
                  'fk INTEGER PRIMARY KEY REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                  ')'
_MAKE_LOG_TABLE = 'CREATE TABLE IF NOT EXISTS logs (' \
                  'log TEXT, ' \
                  'fk INTEGER PRIMARY KEY REFERENCES history(pk) ON DELETE CASCADE ON UPDATE CASCADE' \
                  ')'
_MAKE_SETTINGS_TABLE = 'CREATE TABLE IF NOT EXISTS settings (' \
                       'pk INTEGER PRIMARY KEY NOT NULL, ' \
                       'schema_version INTEGER, ' \
                       'n_keep INTEGER, ' \
                       'n_keep_logs INTEGER)'

_MAKE_HIST_ENTRY = "INSERT INTO history (" \
                   "pk, file, experiment, args, fe_version, train_start, " \
                   "status, n_gpus, n_cpus, n_workers, n_restarts) " \
                   "VALUES (:pk, :fname, :exp, :args, :version, :start, :status, :gpus, :cpus, :workers, 0)"
_MAKE_FEAT_ENTRY = "INSERT INTO features (feature, fk) VALUES (:feature, :fk)"
_MAKE_DS_ENTRY = "INSERT INTO datasets (mode, dataset, fk) VALUES (:mode, :dataset, :fk)"
_MAKE_PIPE_ENTRY = "INSERT INTO pipeline (pipe_op, fk) VALUES (:op, :fk)"
_MAKE_NET_ENTRY = "INSERT INTO network (net_op, fk) VALUES (:op, :fk)"
_MAKE_PP_ENTRY = "INSERT INTO postprocess (pp_op, fk) VALUES (:op, :fk)"
_MAKE_TRACE_ENTRY = "INSERT INTO traces (trace, fk) VALUES (:op, :fk)"
# It would be nice to use a single INSERT ON CONFLICT call here, but Ubuntu 18 apt-get ships version 3.22.0 of sqlite
# and the new syntax wasn't added until 3.24.0
_MAKE_ERR_ENTRY_P1 = "UPDATE errors SET exc_type = :type, exc_tb = :tb WHERE fk = :fk"
_MAKE_ERR_ENTRY_P2 = "INSERT OR IGNORE INTO errors (exc_type, exc_tb, fk) VALUES (:type, :tb, :fk)"
_MAKE_LOG_ENTRY = "INSERT INTO logs (log, fk) VALUES (:log, :fk)"
_MAKE_SETTINGS_ENTRY = "INSERT OR IGNORE INTO settings (pk, schema_version, n_keep, n_keep_logs) " \
                       "VALUES (0, 1, 500, 500)"


def connect(db_path: Optional[str] = None) -> sql.Connection:
    """Open a connection to a sqlite database, creating one if it does not already exist.

    Args:
        db_path: The path to the database file. Or None to default to ~/fastestimator_data/history.db

    Returns:
        An open connection to the database, with schema instantiated and foreign keys enabled.
    """
    if db_path is None:
        db_path = os.path.join(str(Path.home()), 'fastestimator_data', 'history.db')
    if db_path != ':memory:':  # This is a reserved keyword to create an in-memory database
        os.makedirs(os.path.dirname(db_path), exist_ok=True)  # Make sure folders exist before creating disk file
    connection = sql.connect(db_path, detect_types=sql.PARSE_DECLTYPES | sql.PARSE_COLNAMES)
    connection.execute("PRAGMA foreign_keys = 1")  # Enable FK constraints
    connection.row_factory = sql.Row  # Get nice query return objects
    # Build the schema if it doesn't exist
    connection.execute(_MAKE_HIST_TABLE)
    connection.execute(_MAKE_FEAT_TABLE)
    connection.execute(_MAKE_DS_TABLE)
    connection.execute(_MAKE_PIPELINE_TABLE)
    connection.execute(_MAKE_NETWORK_TABLE)
    connection.execute(_MAKE_POST_PROCESS_TABLE)
    connection.execute(_MAKE_TRACE_TABLE)
    connection.execute(_MAKE_ERR_TABLE)
    connection.execute(_MAKE_LOG_TABLE)
    connection.execute(_MAKE_SETTINGS_TABLE)
    connection.execute(_MAKE_SETTINGS_ENTRY)
    connection.commit()
    return connection


def delete(n_keep: int = 20, db_path: Optional[str] = None) -> None:
    """Remove history entries from a database.

    This will also remove associated data such as logs due to foreign key constraints.

    Args:
        n_keep: How many history entries to keep.
        db_path: The path to the database, or None to use the default location.
    """
    db = connect(db_path)
    db.execute(
        "DELETE FROM history WHERE train_start <= ("
        "SELECT train_start FROM history ORDER BY train_start DESC LIMIT 1 OFFSET (?))", [n_keep])
    db.commit()  # Can't vacuum while there are uncommitted changes
    db.execute("VACUUM")  # Free the memory
    db.commit()
    db.close()


def update_settings(n_keep: Optional[int] = None, n_keep_logs: Optional[int] = None,
                    db_path: Optional[str] = None) -> None:
    """Update the history database settings.

    Updated settings will be enforced the next time a training or delete operation is called.

    Args:
        n_keep: How many history entries should be retained.
        n_keep_logs: How many logs should be retained. This value should be <= `n_keep`.
        db_path: The path to the database, or None to use the default location.
    """
    db = connect(db_path)
    # Ensure limits are non-negative
    if n_keep:
        n_keep = max(n_keep, 0)
    if n_keep_logs:
        n_keep_logs = max(n_keep_logs, 0)
    # Perform the update
    if n_keep is not None and n_keep_logs is not None:
        db.execute("UPDATE settings SET n_keep = :keep, n_keep_logs = MIN(:keep, :logs) WHERE pk = 0", {
            'keep': n_keep, 'logs': n_keep_logs
        })
    elif n_keep is not None:
        db.execute("UPDATE settings SET n_keep = :keep, n_keep_logs = MIN(n_keep_logs, :keep) WHERE pk = 0",
                   {'keep': n_keep})
    elif n_keep_logs is not None:
        db.execute("UPDATE settings SET n_keep_logs = MIN(n_keep, (?)) WHERE pk = 0", [n_keep_logs])
    db.commit()
    with closing(db.cursor()) as cursor:
        cursor.execute("SELECT * FROM settings")
        response = from_db_cursor(cursor)
    # Hide implementation details from end user
    response.del_column('pk')
    response.del_column('schema_version')
    print(response)
    db.close()


class HistoryRecorder:
    """A class to record what you're doing.

    This class is intentionally not @traceable.

    It will capture output logs, exceptions, and general information about the training / environment. This class should
    be used as a context manager.

    Args:
        system: The system object corresponding to the current training.
        est_path: The path to the file responsible for creating the current estimator (this is for bookkeeping, it can
            technically be any string).
        db_path: The path to the database, or None to use the default location.
    """
    def __init__(self, system: System, est_path: str, db_path: Optional[str] = None):
        # Prepare db adapters
        sql.register_adapter(bool, int)
        sql.register_converter("BOOL", lambda v: bool(int(v)))
        sql.register_adapter(list, str)
        sql.register_converter("LIST[STR]", lambda v: parse_string_to_python(v))
        # Prepare variables
        self.filename = os.path.basename(est_path)
        self.db_path = db_path if db_path else os.path.join(str(Path.home()), 'fastestimator_data', 'history.db')
        self.system = system
        self.db = None
        self.pk = None
        self.stdout = None

    def __enter__(self) -> None:
        self.db = connect(self.db_path)
        self.pk = self.system.exp_id  # This might be changed later by RestoreWizard. See the _check_for_restart method
        # Check whether an entry for this pk already exists, for example if a user ran .fit() and is now running .test()
        with closing(self.db.cursor()) as cursor:
            exists = cursor.execute("SELECT pk FROM history WHERE pk = (?)", [self.pk])
            exists = exists.fetchall()
        if not exists:
            self.db.execute(
                _MAKE_HIST_ENTRY,
                {
                    'pk': self.pk,
                    'fname': self.filename,
                    'status': 'Launched',
                    'exp': self.system.summary.name,
                    'args': sys.argv[1:],
                    'version': sys.modules['fastestimator'].__version__,
                    'start': datetime.now(),
                    'gpus': torch.cuda.device_count(),
                    'cpus': os.cpu_count(),
                    'workers': self.system.pipeline.num_process
                })
            self.db.executemany(_MAKE_FEAT_ENTRY, self._get_features_in_use())
            self.db.executemany(_MAKE_DS_ENTRY, self._get_datasets_in_use())
            self.db.executemany(_MAKE_PIPE_ENTRY, self._get_items_in_use(self.system.pipeline.ops))
            self.db.executemany(_MAKE_NET_ENTRY, self._get_items_in_use(self.system.network.ops))
            self.db.executemany(_MAKE_PP_ENTRY, self._get_items_in_use(self.system.network.postprocessing))
            self.db.executemany(_MAKE_TRACE_ENTRY, self._get_items_in_use(self.system.traces))
            self.db.execute(_MAKE_LOG_ENTRY, {'log': '', 'fk': self.pk})
            self.db.commit()
        # Take over the output logging
        self.stdout = sys.stdout
        sys.stdout = self

    def __exit__(self, exc_type: Optional[Type], exc_val: Optional[Exception], exc_tb: Optional[Any]) -> None:
        self._check_for_restart()
        self.flush()
        sys.stdout = self.stdout
        # In test mode only overwrite the train_end time if it hasn't already been set
        train_end_query = "(?)" if self.system.mode in ('train', 'eval') else "IFNULL(train_end, (?))"
        query = f"UPDATE history set train_end = {train_end_query}, status = (?) WHERE pk = (?)"
        self.db.execute(
            query,
            [
                datetime.now(),
                "Completed" if exc_type is None else "Aborted" if exc_type == KeyboardInterrupt else "Failed",
                self.pk
            ])
        if exc_type is not None:
            args = {
                'type': exc_type.__name__,
                'tb': "\n".join(traceback.format_exception(exc_type, exc_val, exc_tb)),
                'fk': self.pk
            }
            self.db.execute(_MAKE_ERR_ENTRY_P1, args)
            self.db.execute(_MAKE_ERR_ENTRY_P2, args)
        self.db.commit()
        self._apply_limits()
        self.db.close()

    def _check_for_restart(self) -> None:
        """Determine whether a training has been restarted via RestoreWizard. If so, update the history accordingly.

        If RestoreWizard has been invoked, then the system exp_id will have changed since self.pk was initialized. This
        method will do related bookkeeping, and then swap self.pk for the restored id.
        """
        if self.pk == self.system.exp_id:
            return
        # RestoreWizard reset the system, we are continuing an old training rather than starting a new one
        # First make sure the old entry is still available to edit
        with closing(self.db.cursor()) as cursor:
            exists = cursor.execute("SELECT pk FROM history WHERE pk = (?)", [self.system.exp_id])
            exists = exists.fetchall()
        if exists:
            # If we still have the original entry, we will delete our new one and update the old instead
            self.db.execute("DELETE FROM history WHERE pk = (?)", [self.pk])
        else:
            # The old record doesn't exist, so we will use the new record instead
            self.db.execute("UPDATE history SET pk = (?) WHERE pk = (?)", [self.system.exp_id, self.pk])
        self.pk = self.system.exp_id
        self.db.execute("UPDATE history SET n_restarts = n_restarts + 1 WHERE pk = (?)", [self.pk])
        self.db.commit()

    def _get_features_in_use(self) -> List[Dict[str, str]]:
        """Determine which interesting FE features are being used by the current training.

        Returns:
            A list of entries which can be written into the 'features' db table.
        """
        features = []
        if sys.modules['fastestimator'].fe_deterministic_seed is not None:
            features.append({'feature': 'Deterministic', 'fk': self.pk})
        if any([len(mode_dict) > 1 for mode_dict in self.system.pipeline.data.values()]):
            features.append({'feature': 'MultiDataset', 'fk': self.pk})
        if mixed_precision.global_policy().compute_dtype == 'float16':
            features.append({'feature': 'MixedPrecision', 'fk': self.pk})
        return features

    def _get_datasets_in_use(self) -> List[Dict[str, str]]:
        """Determine which datasets are being used by the current training.

        Returns:
            A list of entries which can be written into the 'datasets' db table.
        """
        datasets = []
        for mode, group in self.system.pipeline.data.items():
            for _, ds in group.items():
                datasets.append({'mode': mode, 'dataset': type(ds).__name__, 'fk': self.pk})
        return datasets

    def _get_items_in_use(self, items: List[Any]) -> List[Dict[str, str]]:
        """Determine which objects are being used by the current training.

        Args:
            items: A list of Schedulers, Ops, and/or traces which are being used by the system.

        Returns:
            The elements from `items` converted into database-ready entries.
        """
        ops = []
        for op in items:
            op_list = [op]
            if isinstance(op, Scheduler):
                op_list = list(filter(lambda x: x is not None, op.get_all_values()))
                op_list.append(op)  # Put scheduler in too so that usage can be tracked too
            ops.extend([{'op': type(elem).__name__, 'fk': self.pk} for elem in op_list])
        return ops

    def _apply_limits(self) -> None:
        """Remove old history and/or log entries if they exceed the limits defined in the settings table.
        """
        self.db.execute("DELETE FROM history WHERE train_start <= ("
                        "SELECT train_start FROM history ORDER BY train_start DESC LIMIT 1 OFFSET ("
                        "SELECT n_keep FROM settings WHERE pk = 0))")
        self.db.execute("DELETE FROM logs WHERE fk IN ("
                        "SELECT pk FROM history ORDER BY train_start DESC LIMIT 1 OFFSET ("
                        "SELECT n_keep_logs FROM settings WHERE pk = 0))")
        self.db.commit()  # Have to commit before vacuuming
        if sum(int(digit) for digit in str(abs(self.pk))) % 10 == 0:
            # 10% of time do a vacuum (expensive). We don't use random.randint here due to deterministic training. Also,
            # don't use pk directly because last digit is not uniformly distributed.
            self.db.execute("PRAGMA VACUUM;")
            self.db.commit()
        else:
            # Otherwise do a less costly optimize
            self.db.execute("PRAGMA optimize;")
            self.db.commit()
        self.db.close()

    def write(self, output: str) -> None:
        self._check_for_restart()  # Check here instead of just waiting for __exit__ in case system powers off later
        self.stdout.write(output)
        if multiprocessing.current_process().name == 'MainProcess':
            # Flush can also get invoked by pipeline multi-processing, but db should only be accessed by main thread.
            # This can happen, for example, when pipeline prints a warning that a certain key is unused and will be
            # dropped.
            self.db.execute('UPDATE logs SET log = log || (?) WHERE fk = (?)', [output, self.pk])
            self.db.commit()

    def flush(self) -> None:
        self.stdout.flush()


class HistoryReader:
    """A class to read history information from the database.

    This class is intentionally not @traceable.

    This class should be used as as a context manager, for example:

    ```python
    with HistoryReader() as reader:
        reader.read_basic()
    ```

    Args:
        db_path: The path to the database, or None to use the default location.
    """
    def __init__(self, db_path: Optional[str] = None):
        self.db_path = db_path
        self.db = None  # sql.Connection
        self.response = None  # List[sql.Row]

    def __enter__(self) -> 'HistoryReader':
        self.db = connect(self.db_path)
        self.db.set_trace_callback(print)  # Show the query in case user wants to adapt it later
        return self

    def __exit__(self, exc_type: Optional[Type], exc_val: Optional[Exception], exc_tb: Optional[Any]) -> None:
        self.db.close()

    def read_basic(self,
                   limit: int = 10,
                   interactive: bool = False,
                   include_args: bool = False,
                   errors: bool = False,
                   include_pk: bool = False,
                   include_features: bool = False,
                   include_traces: bool = False,
                   include_datasets: bool = False,
                   include_pipeline: bool = False,
                   include_network: bool = False,
                   as_csv: bool = False) -> None:
        """Perform a pre-defined (and possibly interactive) set of sql selects against the history database.

        Outputs will be printed to stdout.

        Args:
            limit: The maximum number of responses to look up.
            interactive: Whether to run this function interactively, prompting the user for additional input along the
                way. This enables things like error and log retrieval for individual experiments.
            include_args: Whether to output the arguments used to run each experiment.
            errors: Whether to filter the output to only include failed experiments, as well as including more
                information about the specific errors that occurred.
            include_pk: Whether to include the primary keys (experiment ids) of each history entry.
            include_features: Whether to include the FE features that were employed by each training.
            include_traces: Whether to include the traces that were used in each training.
            include_datasets: Whether to include the dataset (classes) that were used in each training.
            include_pipeline: Whether to include the pipeline ops that were used in each training.
            include_network: Whether to include the network (post)processing ops that were used in each training.
            as_csv: Whether to print the output as a csv rather than in a formatted table.
        """
        # Build the query string
        error_select = ", errors.exc_type error" if errors else ''
        error_join = "LEFT JOIN errors ON errors.fk = h.pk " if errors else ''
        error_where = " WHERE h.status <> 'Completed' " if errors else ''
        feature_select = ", fg.features" if include_features else ''
        feature_join = "LEFT JOIN (" \
                       "SELECT fk, GROUP_CONCAT(feature, ', ') features FROM features f GROUP BY f.fk" \
                       ") fg ON fg.fk = h.pk " if include_features else ''
        dataset_select = ", dsg.datasets " if include_datasets else ''
        dataset_join = "LEFT JOIN (" \
                       "SELECT fk, GROUP_CONCAT(dataset || ' (' || mode || ')', ', ') datasets " \
                       "FROM datasets ds GROUP BY ds.fk" \
                       ") dsg ON dsg.fk = h.pk " if include_datasets else ''
        pipeline_select = ", pg.pipeline_ops" if include_pipeline else ''
        pipeline_join = "LEFT JOIN (" \
                        "SELECT fk, GROUP_CONCAT(pipe_op, ', ') pipeline_ops FROM pipeline p GROUP BY p.fk" \
                        ") pg ON pg.fk = h.pk " if include_pipeline else ''
        network_select = ", ng.network_ops, ppg.postprocessing_ops" if include_network else ''
        network_join = "LEFT JOIN (" \
                       "SELECT fk, GROUP_CONCAT(net_op, ', ') network_ops FROM network n GROUP BY n.fk" \
                       ") ng ON ng.fk = h.pk " \
                       "LEFT JOIN (" \
                       "SELECT fk, GROUP_CONCAT(pp_op, ', ') postprocessing_ops FROM postprocess pp GROUP BY pp.fk" \
                       ") ppg ON ppg.fk = h.pk " if include_network else ''
        trace_select = ", tg.traces " if include_traces else ''
        trace_join = "LEFT JOIN (" \
                     "SELECT fk, GROUP_CONCAT(trace, ', ') traces FROM traces t GROUP BY t.fk" \
                     ") tg ON tg.fk = h.pk " if include_traces else ''
        query = f"SELECT h.*{error_select}{feature_select}{dataset_select}{pipeline_select}{network_select}" \
                f"{trace_select} FROM history h {error_join}{feature_join}{dataset_join}{pipeline_join}{network_join}" \
                f"{trace_join}{error_where}ORDER BY h.train_start DESC LIMIT (?)"
        # We have to hide these after-the-fact since later process may require pk behind the scenes
        hide = []
        if not include_pk:
            hide.append('pk')
        if not include_args:
            hide.append('args')
        self.read_sql(query, args=[limit], hide_cols=hide, as_csv=as_csv, interactive=interactive)

    def read_sql(self,
                 query: str,
                 args: Iterable[Any] = (),
                 hide_cols: Iterable[str] = (),
                 as_csv: bool = False,
                 interactive: bool = False) -> None:
        """Perform a (possibly interactive) sql query against the database.

        Args:
            query: The sql query to execute.
            args: Any parameterized arguments to be inserted into the `query`.
            hide_cols: Any columns to hide from the printed output.
            as_csv: Whether to print the output in csv format or in table format.
            interactive: Whether to run this function interactively, prompting the user for additional input along the
                way. This enables things like error and log retrieval for individual experiments.
        """
        with closing(self.db.cursor()) as cursor:
            cursor.execute(query, args)
            self.response = cursor.fetchall()
            names = [col[0] for col in cursor.description]
        # Build nice output table
        table = PrettyTable()
        table.field_names = names
        for row in self.response:
            table.add_row(row)
        for col in hide_cols:
            if col in table.field_names:
                table.del_column(col)
        if interactive:
            table.add_autoindex()
        if as_csv:
            print(table.get_csv_string())
        else:
            print(table)
        if interactive:
            while True:
                inp = input("\033[93m{}\033[00m".format("Enter --help for available command details. Enter without an "
                                                        "argument to re-print the current response. X to exit.\n"))
                if inp in ('X', 'x'):
                    break
                if inp == "":
                    print(query)
                    print(table)
                    continue
                new_query = self._parse_input(inp)
                if new_query:
                    return self.read_sql(new_query, hide_cols=hide_cols, as_csv=as_csv, interactive=interactive)

    def _parse_input(self, inp: str) -> Optional[str]:
        """Take cli input and run it through command parsers to execute an appropriate subroutine.

        Args:
            inp: The cli input provided by an end user.

        Returns:
            The output (if any) of the appropriate sub-command after executing on the given input.
        """
        parser = argparse.ArgumentParser(allow_abbrev=False)
        subparsers = parser.add_subparsers()
        subparsers.required = True
        subparsers.dest = 'cmd'
        self._configure_sql_parser(subparsers)
        self._configure_log_parser(subparsers)
        self._configure_err_parser(subparsers)
        self._configure_vis_parser(subparsers)
        try:
            args, unknown = parser.parse_known_args(inp.split())
        except SystemExit:
            return
        return args.func(vars(args), unknown)

    def _configure_sql_parser(self, subparsers: argparse._SubParsersAction) -> None:
        """Add a sql parser to an existing argparser.

        Args:
            subparsers: The parser object to be appended to.
        """
        p_sql = subparsers.add_parser('sql',
                                      description='Provide a new sql query to be executed',
                                      formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      allow_abbrev=False)
        p_sql.add_argument('query', metavar='<Query>', type=str, nargs='+', help="ex: sql SELECT * FROM history")
        p_sql.set_defaults(func=self._echo_sql)

    def _configure_log_parser(self, subparsers: argparse._SubParsersAction) -> None:
        """Add a log parser to an existing argparser.

        Args:
            subparsers: The parser object to be appended to.
        """
        p_log = subparsers.add_parser(
            'log',
            description='Retrieve one or more output logs. This command requires '
            'that you currently have experiments selected.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            allow_abbrev=False)
        p_log.add_argument('indices',
                           metavar='I',
                           type=int,
                           nargs='+',
                           help="Indices of experiments for which to print logs")
        p_log.add_argument(
            '--file',
            metavar='F',
            action=SaveAction,
            default=False,
            dest='file',
            nargs='?',
            help='Whether to write the logs to disk. May be accompanied by a directory or filename into which to save \
                 the log(s). If none is specified then the ~/fastestimator_data directory will be used.')
        p_log.set_defaults(func=self._fetch_logs)

    def _configure_err_parser(self, subparsers: argparse._SubParsersAction) -> None:
        """Add an error parser to an existing argparser.

        Args:
            subparsers: The parser object to be appended to.
        """
        p_err = subparsers.add_parser(
            'err',
            description='Retrieve one or more error tracebacks. This command requires '
            'that you currently have experiments selected.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            allow_abbrev=False)
        p_err.add_argument('indices',
                           metavar='I',
                           type=int,
                           nargs='+',
                           help="Indices of experiments for which to print error tracebacks")
        p_err.set_defaults(func=self._fetch_errs)

    def _configure_vis_parser(self, subparsers: argparse._SubParsersAction) -> None:
        """Add a visualization parser to an existing argparser.

        Args:
            subparsers: The parser object to be appended to.
        """
        p_vis = subparsers.add_parser(
            'vis',
            description='Visualize logs for one or more experiments. This command requires '
            'that you currently have experiments selected.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            allow_abbrev=False)
        p_vis.add_argument('indices',
                           metavar='idx',
                           type=int,
                           nargs='*',
                           help="Indices of experiments for which to print logs")
        group = p_vis.add_mutually_exclusive_group()
        group.add_argument('--ignore',
                           metavar='I',
                           type=str,
                           nargs='+',
                           help="The names of metrics to ignore though they may be present in the log files")
        group.add_argument('--include',
                           metavar='Y',
                           type=str,
                           nargs='+',
                           help="The names of metrics to include. If provided, any other metrics will be ignored.")
        p_vis.add_argument('--smooth',
                           metavar='<float>',
                           type=float,
                           help="The amount of gaussian smoothing to apply (zero for no smoothing)",
                           default=1)
        p_vis.add_argument('--pretty_names', help="Clean up the metric names for display", action='store_true')
        p_vis.add_argument('-g', '--group', dest='groups', default={}, action=_GroupAction, nargs="*")

        legend_group = p_vis.add_argument_group('legend arguments')
        legend_x_group = legend_group.add_mutually_exclusive_group(required=False)
        legend_x_group.add_argument('--common_legend',
                                    dest='share_legend',
                                    help="Generate one legend total",
                                    action='store_true',
                                    default=True)
        legend_x_group.add_argument('--split_legend',
                                    dest='share_legend',
                                    help="Generate one legend per graph",
                                    action='store_false',
                                    default=False)

        save_group = p_vis.add_argument_group('output arguments')
        save_x_group = save_group.add_mutually_exclusive_group(required=False)
        save_x_group.add_argument(
            '--save',
            nargs='?',
            metavar='<Save Dir>',
            dest='save',
            action=SaveAction,
            default=False,
            help="Save the output image. May be accompanied by a directory into \
                  which the file is saved. If no output directory is specified, the history directory will be used")
        save_x_group.add_argument('--display',
                                  dest='save',
                                  action='store_false',
                                  help="Render the image to the UI (rather than saving it)",
                                  default=True)
        save_x_group.set_defaults(save_dir=None)
        p_vis.set_defaults(func=self._vis_logs)

    @staticmethod
    def _echo_sql(args: Dict[str, Any], unknown: List[str]) -> Optional[str]:
        """A method to compile parsed user input back into a single sql query.

        Args:
            args: The CLI arguments provided by the user.
            unknown: Any CLI arguments not matching known inputs.

        Returns:
            A single string containing the user sql query.
        """
        if len(unknown) > 0:
            print("unrecognized arguments: ", str.join(", ", unknown))
            return None
        return " ".join(args['query'])

    def _fetch_logs(self, args: Dict[str, Any], unknown: List[str]) -> None:
        """A method to collect and return a given set of logs from the database.

        Args:
            args: The CLI arguments provided by the user.
            unknown: Any CLI arguments not matching known inputs.
        """
        if len(unknown) > 0:
            print("unrecognized arguments: ", str.join(", ", unknown))
            return
        save = args['file']
        save_path = None
        if save:
            save_path = args['file_dir']
            if save_path is None:
                save_path = os.path.join(str(Path.home()), 'fastestimator_data')
                save = 'dir'
                print(f"Writing log(s) to {save_path}")
            else:
                save = 'file'
                print(f'Writing log to {save_path}')
        logs = {}
        for idx in args['indices']:
            selection = self.response[idx - 1]  # Auto index starts at 1
            pk = selection['pk']
            with closing(self.db.cursor()) as cursor:
                cursor.execute("SELECT log FROM logs WHERE logs.fk = (?)", [pk])
                logs[idx] = cursor.fetchall()
        with open(save_path, 'w') if save == 'file' else NonContext() as f:
            f = sys.stdout if f is None else f
            for idx, log in logs.items():
                with open(os.path.join(save_path, f"{idx}.txt"), 'w') if save == 'dir' else NonContext() as f1:
                    f1 = f if f1 is None else f1
                    if log:
                        f1.write(f'\n@@@@@@@@@@@ Log for Index {idx} @@@@@@@@@@@\n\n')
                        f1.write(log[0]['log'])
                        f1.write('\n')
                    else:
                        f1.write(f"No logs found for Index {idx}\n")

    def _fetch_errs(self, args: Dict[str, Any], unknown: List[str]) -> None:
        """A method to collect and return a given set of error logs from the database.

        Args:
            args: The CLI arguments provided by the user.
            unknown: Any CLI arguments not matching known inputs.
        """
        if len(unknown) > 0:
            print("unrecognized arguments: ", str.join(", ", unknown))
            return
        for idx in args['indices']:
            selection = self.response[idx - 1]  # Auto index starts at 1
            pk = selection['pk']
            with closing(self.db.cursor()) as cursor:
                cursor.execute("SELECT exc_tb FROM errors WHERE errors.fk = (?)", [pk])
                err = cursor.fetchall()
            if err:
                print(f'@@@@@@@@@@@ Traceback for Index {idx} @@@@@@@@@@@')
                print(err[0]['exc_tb'])
            else:
                print(f"No error traceback found for Index {idx}")

    def _vis_logs(self, args: Dict[str, Any], unknown: List[str]) -> None:
        """A method to collect and visualize a given set of logs from the database.

        Args:
            args: The CLI arguments provided by the user.
            unknown: Any CLI arguments not matching known inputs.
        """
        if len(unknown) > 0:
            print("unrecognized arguments: ", str.join(", ", unknown))
            return
        save_dir = args['save_dir']
        if args['save'] and save_dir is None:
            save_dir = os.path.join(str(Path.home()), 'fastestimator_data', 'logs.png')
        group_indices = [x for y in args['groups'].values() for x in y]
        pks = {idx: self.response[idx - 1]['pk'] for idx in set(args['indices'] + group_indices)}
        if len(pks) == 0:
            return
        with closing(self.db.cursor()) as cursor:
            cursor.execute(
                "SELECT H.pk, H.experiment, L.log "
                "FROM logs L LEFT JOIN history H ON L.fk = H.pk "
                "WHERE L.fk IN ({seq})".format(seq=','.join(['?'] * len(pks))),
                list(pks.values()))
            logs = cursor.fetchall()
        logs = {elem['pk']: elem for elem in logs}
        failures = 0
        for idx, pk in pks.items():
            if pk not in logs:
                print(f"No logs found for Index {idx}")
                failures += 1
        if failures:
            return
        groups = defaultdict(list)  # {group_name: [experiment(s)]}
        for idx, pk in pks.items():
            log = logs[pk]
            experiment = parse_log_iter(
                log['log'].split('\n'),
                Summary(str(idx) if log['experiment'] is None else f"{log['experiment']} ({idx})"))
            if idx in args['indices']:
                groups[experiment.name].append(experiment)
            for group_name, group_indices in args['groups'].items():
                if idx in group_indices:
                    groups[group_name].append(experiment)
        experiments = [average_summaries(name, exps) for name, exps in groups.items()]
        visualize_logs(experiments,
                       save_path=save_dir,
                       smooth_factor=args['smooth'],
                       share_legend=args['share_legend'],
                       pretty_names=args['pretty_names'],
                       ignore_metrics=args['ignore'],
                       include_metrics=args['include'])


class _GroupAction(argparse._AppendAction):
    """An argparse action which can be invoked multiple times in order to build a dictionary of entries.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) < 2:
            raise argparse.ArgumentError(self, "--group arguments should of the form <name> <idx1> [<idx2> ...]")
        key = values[0]
        indices = [int(val) for val in values[1:]]
        items = getattr(namespace, self.dest, None)
        if items is None:
            items = {}
        items = items.copy()
        items[key] = indices
        setattr(namespace, self.dest, items)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import fastestimator as fe
from fastestimator.trace import Trace


class TestTraceSortTraces(unittest.TestCase):
    """ This test has dependency on:
    * fe.trace.Trace
    * fe.util.util.to_set()
    """
    def test_estimator_sort_traces_with_available_outputs_have_dependency(self):
        trace1 = Trace(inputs="x", outputs="y")
        trace2 = Trace(inputs="y", outputs="z")
        trace3 = Trace(inputs="z", outputs="w")

        sorted_traces = fe.trace.sort_traces([trace3, trace2, trace1], ds_ids=[], available_outputs="x")
        self.assertEqual(sorted_traces, [trace1, trace2, trace3])

    def test_estimator_sort_traces_with_available_outputs_have_no_dependency(self):
        trace1 = Trace(inputs="x", outputs="y")
        trace2 = Trace(inputs="x", outputs="z")
        trace3 = Trace(inputs="x", outputs="x")

        with self.subTest("[trace1, trace2]"):
            sorted_traces = fe.trace.sort_traces([trace1, trace2], ds_ids=[], available_outputs="x")
            self.assertEqual(sorted_traces, [trace1, trace2])

        with self.subTest("[trace2, trace1]"):
            sorted_traces = fe.trace.sort_traces([trace2, trace1], ds_ids=[], available_outputs="x")
            self.assertEqual(sorted_traces, [trace2, trace1])

        with self.subTest("[trace1, trace3]"):
            sorted_traces = fe.trace.sort_traces([trace1, trace3], ds_ids=[], available_outputs="x")
            self.assertEqual(sorted_traces, [trace1, trace3])

        with self.subTest("[trace3, trace1]"):
            sorted_traces = fe.trace.sort_traces([trace3, trace1], ds_ids=[], available_outputs="x")
            self.assertEqual(sorted_traces, [trace3, trace1])

    def test_estimator_sort_traces_with_available_outputs_no_input_match(self):
        trace1 = Trace(inputs="y", outputs="z")
        with self.assertRaises(AssertionError):
            sorted_traces = fe.trace.sort_traces([trace1], ds_ids=[], available_outputs="x")

    def test_estimator_sort_traces_without_available_outputs_have_dependency(self):
        trace1 = Trace(inputs="x", outputs="y")
        trace2 = Trace(inputs="y", outputs="z")
        trace3 = Trace(inputs="z", outputs="w")

        sorted_traces = fe.trace.sort_traces([trace3, trace2, trace1], ds_ids=[])
        self.assertEqual(sorted_traces, [trace1, trace2, trace3])

    def test_estimator_sort_traces_without_available_outputs_have_no_dependency(self):
        trace1 = Trace(inputs="x", outputs="y")
        trace2 = Trace(inputs="a", outputs="b")

        with self.subTest("[trace1, trace2]"):
            sorted_traces = fe.trace.sort_traces([trace1, trace2], ds_ids=[])
            self.assertEqual(sorted_traces, [trace1, trace2])

        with self.subTest("[trace2, trace1]"):
            sorted_traces = fe.trace.sort_traces([trace2, trace1], ds_ids=[])
            self.assertEqual(sorted_traces, [trace2, trace1])

    def test_estimator_sort_traces_cycle_dependency(self):
        trace1 = Trace(inputs="x", outputs="y")
        trace2 = Trace(inputs="y", outputs="z")
        trace3 = Trace(inputs="z", outputs="x")
        trace4 = Trace(inputs="x", outputs="a")

        with self.subTest("available_outputs='x'"):
            sorted_traces = fe.trace.sort_traces([trace1, trace2, trace3, trace4], ds_ids=[], available_outputs="y")
            self.assertEqual(sorted_traces, [trace2, trace3, trace4, trace1])

        with self.subTest("available_output=None"):
            with self.assertRaises(AssertionError):
                sorted_traces = fe.trace.sort_traces([trace1, trace2, trace3, trace4], ds_ids=[])


class TestTraceNames(unittest.TestCase):
    def test_forbidden_names_semi(self):
        with self.assertRaises(AssertionError):
            fe.trace.Trace(inputs="x:")
        with self.assertRaises(AssertionError):
            fe.trace.Trace(outputs="x:")

    def test_forbidden_names_colon(self):
        with self.assertRaises(AssertionError):
            fe.trace.Trace(inputs="x;")
        with self.assertRaises(AssertionError):
            fe.trace.Trace(outputs="x;")



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import csv
import os
import tempfile
import unittest
from collections import defaultdict

from fastestimator.test.unittest_util import is_equal, sample_system_object
from fastestimator.trace.io import CSVLogger
from fastestimator.util.data import Data


class TestCSVLogger(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = Data({'mode': 'train', 'epoch': 1})
        cls.dict = defaultdict(list, {'mode': ['train'], 'epoch': [0]})
        cls.csv_path = os.path.join(tempfile.mkdtemp(), 'test_csv_logger.csv')

    def test_on_begin(self):
        csvlogger = CSVLogger(filename=self.csv_path)
        csvlogger.on_begin(data=self.data)
        self.assertTrue(is_equal(csvlogger.data, defaultdict(list)))

    def test_on_epoch_end(self):
        csvlogger = CSVLogger(filename=self.csv_path)
        csvlogger.system = sample_system_object()
        csvlogger.data = defaultdict(list)
        csvlogger.on_epoch_end(data=self.data)
        self.assertDictEqual(csvlogger.data, self.dict)

    def test_on_end(self):
        csvlogger = CSVLogger(filename=self.csv_path)
        csvlogger.system = sample_system_object()
        csvlogger.data = defaultdict(list, {'mode': ['train'], 'epoch': [0]})
        # remove csv file previously created if it exists
        if os.path.exists(self.csv_path):
            os.remove(self.csv_path)
        csv_data = []
        # call on_end here and create new csv
        csvlogger.on_end(data=self.data)
        with open(self.csv_path) as f:
            records = csv.DictReader(f)
            for param in records:
                csv_data.append(defaultdict(list, dict(param)))
        with self.subTest('Check if path exists'):
            self.assertTrue(os.path.exists(self.csv_path))
        with self.subTest('Check mode value stored in csv file'):
            self.assertEqual(csv_data[0]['mode'], csvlogger.data['mode'][0])
        with self.subTest('Check epoch value stored in csv file'):
            self.assertEqual(int(csv_data[0]['epoch']), csvlogger.data['epoch'][0])



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
from typing import Optional, TypeVar

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend.maximum import maximum
from fastestimator.backend.reduce_sum import reduce_sum
from fastestimator.backend.reshape import reshape
from fastestimator.backend.tensor_pow import tensor_pow
from fastestimator.backend.to_tensor import to_tensor
from fastestimator.util.util import TENSOR_TO_NP_DTYPE

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def iwd(tensor: Tensor,
        power: float = 1.0,
        max_prob: float = 0.95,
        pairwise_distance: float = 1.0,
        eps: Optional[Tensor] = None) -> Tensor:
    """Compute the Inverse Weighted Distance from the given input.

    This can be used as an activation function for the final layer of a neural network instead of softmax. For example,
    instead of: model.add(layers.Dense(classes, activation='softmax')), you could use:
    model.add(layers.Dense(classes, activation=lambda x: iwd(tf.nn.sigmoid(x))))

    This method can be used with Numpy data:
    ```python
    n = np.array([[0.5]*5, [0]+[1]*4])
    b = fe.backend.iwd(n)  # [[0.2, 0.2, 0.2, 0.2, 0.2], [0.95, 0.0125, 0.0125, 0.0125, 0.0125]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([[0.5]*5, [0]+[1]*4])
    b = fe.backend.iwd(n)  # [[0.2, 0.2, 0.2, 0.2, 0.2], [0.95, 0.0125, 0.0125, 0.0125, 0.0125]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([[0.5]*5, [0]+[1]*4])
    b = fe.backend.iwd(n)  # [[0.2, 0.2, 0.2, 0.2, 0.2], [0.95, 0.0125, 0.0125, 0.0125, 0.0125]]
    ```

    Args:
        tensor: The input value. Should be of shape (Batch, C) where every element in C corresponds to a (non-negative)
            distance to a target class.
        power: The power to raise the inverse distances to. 1.0 results in a fairly intuitive probability output. Larger
            powers can widen regions of certainty, whereas values between 0 and 1 can widen regions of uncertainty.
        max_prob: The maximum probability to assign to a class estimate when it is distance zero away from the target.
            For numerical stability this must be less than 1.0. We have found that using smaller values like 0.95 can
            lead to natural adversarial robustness.
        pairwise_distance: The distance to any other class when the distance to a target class is zero. For example, if
            you have a perfect match for class 'a', what distance should be reported to class 'b'. If you have a metric
            where this isn't constant, just use an approximate expected distance. In that case `max_prob` will only give
            you approximate control over the true maximum probability.
        eps: The numeric stability constant to be used when d approaches zero. If None then it will be computed using
            `max_prob` and `pairwise_distance`. If not None, then `max_prob` and `pairwise_distance` will be ignored.

    Returns:
        A probability distribution of shape (Batch, C) where smaller distances from `tensor` correspond to larger
        probabilities.
    """
    if eps is None:
        eps = np.array(pairwise_distance * math.pow((1.0 - max_prob) / (max_prob * (tensor.shape[-1] - 1)), 1 / power),
                       dtype=TENSOR_TO_NP_DTYPE[tensor.dtype])
        eps = to_tensor(
            eps, target_type='torch' if isinstance(tensor, torch.Tensor) else 'tf' if tf.is_tensor(tensor) else 'np')
        if isinstance(eps, torch.Tensor):
            eps = eps.to("cuda:0" if torch.cuda.is_available() else "cpu")
    tensor = maximum(tensor, eps)
    tensor = tensor_pow(1.0 / tensor, power)
    tensor = tensor / reshape(reduce_sum(tensor, axis=-1), shape=[-1, 1])
    return tensor
